"""
This Python 3.11 module implements the rHEALPix map projection.

- Alexander Raichev (AR), 2013-01-26: Refactored code from release 0.3.

NOTE:

All lengths are measured in meters and all angles are measured in radians
unless indicated otherwise.
By 'ellipsoid' below, I mean an oblate ellipsoid of revolution.
"""

# *****************************************************************************
#       Copyright (C) 2013 Alexander Raichev <alex.raichev@gmail.com>
#
#  Distributed under the terms of the GNU Lesser General Public License (LGPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************

# Import third-party modules.
from numpy import pi, sign, array, identity, dot, deg2rad, rad2deg
from typing import Callable

# Import my modules.
from vgrid.utils.rhealpixdggs.pj_healpix import (
    healpix_sphere,
    healpix_sphere_inverse,
    healpix_ellipsoid,
    healpix_ellipsoid_inverse,
)
from vgrid.utils.rhealpixdggs.utils import auth_rad

# Matrix for anticlockwise rotation by pi/2:
ROTATE1 = array([[0, -1], [1, 0]])
# Matrix for anticlockwise rotation by pi:
ROTATE2 = dot(ROTATE1, ROTATE1)
# Matrix for anticlockwise rotation by 3*pi/2.
ROTATE3 = dot(ROTATE2, ROTATE1)
# Dictionary of all powers of ROTATE1 and its inverse (ROTATE3).
ROTATE = {
    0: identity(2, int),
    1: ROTATE1,
    2: ROTATE2,
    3: ROTATE3,
    -1: ROTATE3,
    -2: ROTATE2,
    -3: ROTATE1,
}


def combine_triangles(
    x: float,
    y: float,
    north_square: int = 0,
    south_square: int = 0,
    inverse: bool = False,
) -> tuple[float, float]:
    """
    Rearrange point `(x, y)` in the HEALPix projection by
    combining the polar triangles into two polar squares.
    Put the north polar square in position `north_square` and
    the south polar square in position `south_square`.
    If `inverse` = True, uncombine the polar triangles.

    INPUT:

    - `x, y` - Coordinates in the HEALPix projection of the unit sphere.
    - `north_square, south_square` - Integers between 0 and 3 indicating
      the positions of the north_square polar square and south_square polar
      square respectively.
      See rhealpix_sphere() docstring for a diagram.
    - `inverse` - (Optional; default = False) Boolean. If False, then compute
      forward function. If True, then compute inverse function.

    EXAMPLES::

        >>> u, v = -pi/4, pi/3
        >>> x, y = combine_triangles(u, v)
        >>> print(tuple(x.tolist() for x in my_round((x, y), 15)))
        (-1.832595714594046, 1.570796326794896)
        >>> print(tuple(x.tolist() for x in my_round(combine_triangles(x, y, inverse=True), 15)))
        (-0.785398163397448, 1.047197551196598)
        >>> print(my_round((u, v), 15))
        (-0.785398163397448, 1.047197551196598)

    """
    # Ensure north_square and south_square lie in {0, 1, 2, 3}.
    north_square = north_square % 4
    south_square = south_square % 4
    c, region = triangle(
        x, y, north_square=north_square, south_square=south_square, inverse=inverse
    )
    if region == "equatorial":
        # (x,y) remains fixed
        return x, y
    xy = array((x, y))
    tc = array((-3 * pi / 4 + c * pi / 2, sign(y) * pi / 2))
    if not inverse:
        # Forward function.
        # Rotate (x, y) about tc and then translate it to
        # the tip u of the polar triangle it will be assembled upon.
        if region == "north_polar":
            u = array((-3 * pi / 4 + north_square * pi / 2, pi / 2))
            x, y = dot(ROTATE[c - north_square], xy - tc) + u
        elif region == "south_polar":
            u = array((-3 * pi / 4 + south_square * pi / 2, -pi / 2))
            x, y = dot(ROTATE[-(c - south_square)], xy - tc) + u
    else:
        # Inverse function.
        # Unrotate (x, y) about u and then translate it to tc.
        if region == "north_polar":
            u = array((-3 * pi / 4 + north_square * pi / 2, pi / 2))
            x, y = dot(ROTATE[-(c - north_square)], xy - u) + tc
        elif region == "south_polar":
            u = array((-3 * pi / 4 + south_square * pi / 2, -pi / 2))
            x, y = dot(ROTATE[c - south_square], xy - u) + tc
    return x, y


def triangle(
    x: float,
    y: float,
    north_square: int = 0,
    south_square: int = 0,
    inverse: bool = False,
) -> tuple[int, str]:
    """
    Return the number of the polar triangle and region that `(x, y)` lies in.
    If `inverse` = False, then assume `(x,y)` lies in the image of the HEALPix
    projection of the unit sphere.
    If `inverse` = True, then assume `(x,y)` lies in the image of the
    `(north_square, south_square)`-rHEALPix projection of the unit sphere.

    INPUT:

    - `x, y` - Coordinates in the HEALPix or rHEALPix (if `inverse` = True)
      projection of the unit sphere.
    - `north_square, south_square` - Integers between 0 and 3 indicating the
      positions of the north_square pole square and south_square pole square
      respectively.
      See rhealpix_sphere() docstring for a diagram.
    - `inverse` - (Optional; default = False) Boolean. If False, then compute
      forward function. If True, then compute inverse function.

    OUTPUT:

    The pair (triangle_number, region).
    Here region equals 'north_polar' (polar), 'south_polar' (polar), or
    'equatorial', indicating where `(x, y)` lies.
    If region = 'equatorial', then triangle_number = None.
    Suppose now that region != 'equatorial'.
    If `inverse` = False, then triangle_number is the number (0, 1, 2, or 3) of
    the HEALPix polar triangle Z that `(x, y)` lies in.
    If `inverse` = True, then triangle_number is the number (0, 1, 2, or 3) of
    the HEALPix polar triangle that `(x, y)` will get moved into.

    EXAMPLES::

        >>> triangle(-pi/4, pi/4 + 0.1)
        (1, 'north_polar')
        >>> triangle(-3*pi/4 + 0.1, pi/2, inverse=True)
        (1, 'north_polar')

    NOTES:

    In the HEALPix projection, the polar triangles are labeled 0--3 from
    east to west like this::

            *       *       *       *
          * 0 *   * 1 *   * 2 *   * 3 *
        *-------*-------*-------*-------*
        |       |       |       |       |
        |       |       |       |       |
        |       |       |       |       |
        *-------*-------*-------*-------*
          * 0 *   * 1 *   * 2 *   * 3 *
            *       *       *       *

    In the rHEALPix projection these polar triangles get rearranged
    into a square with the triangles numbered `north_square` and `south_square`
    remaining fixed.
    For example, if `north_square` = 1 and `south_square` = 3,
    then the triangles get rearranged this way::

        North polar square:     *-------*
                                | * 3 * |
                                | 0 * 2 |
                                | * 1 * |
                            ----*-------*----

        South polar square: ----*-------*----
                                | * 3 * |
                                | 2 * 0 |
                                | * 1 * |
                                *-------*

    """
    if not inverse:
        # Forward function.
        # Find the region (x, y) lies in.
        if y > pi / 4:
            region = "north_polar"
        elif y < -pi / 4:
            region = "south_polar"
        else:
            region = "equatorial"
        # Find the triangle number of (x, y) in the image of the
        # HEALPix projection.
        if region == "equatorial":
            triangle_number = None
        else:
            if x < -pi / 2:
                triangle_number = 0
            elif x >= -pi / 2 and x < 0:
                triangle_number = 1
            elif x >= 0 and x < pi / 2:
                triangle_number = 2
            else:
                triangle_number = 3
    else:
        # Inverse function.
        # Find the region (x, y) lies in.
        if y > pi / 4:
            region = "north_polar"
        elif y < -pi / 4:
            region = "south_polar"
        else:
            region = "equatorial"
        # Find HEALPix polar triangle number that (x, y) moves to when
        # the rHEALPix polar square is disassembled.
        eps = 1e-15  # Fuzz to avoid some rounding errors.
        if region == "equatorial":
            triangle_number = None
        elif region == "north_polar":
            L1 = x - (-3 * pi / 4 + (north_square - 1) * pi / 2)
            L2 = -x + (-3 * pi / 4 + (north_square + 1) * pi / 2)
            if y < L1 - eps and y >= L2 - eps:
                triangle_number = (north_square + 1) % 4
            elif y >= L1 - eps and y > L2 + eps:
                triangle_number = (north_square + 2) % 4
            elif y > L1 + eps and y <= L2 + eps:
                triangle_number = (north_square + 3) % 4
            else:
                triangle_number = north_square
        else:
            # region == 'south_polar':
            L1 = x - (-3 * pi / 4 + (south_square + 1) * pi / 2)
            L2 = -x + (-3 * pi / 4 + (south_square - 1) * pi / 2)
            if y <= L1 + eps and y > L2 + eps:
                triangle_number = (south_square + 1) % 4
            elif y < L1 - eps and y <= L2 + eps:
                triangle_number = (south_square + 2) % 4
            elif y >= L1 - eps and y < L2 - eps:
                triangle_number = (south_square + 3) % 4
            else:
                triangle_number = south_square
    return triangle_number, region


def rhealpix_sphere(
    lam: float,
    phi: float,
    north_square: int = 0,
    south_square: int = 0,
    region: str = "none",
) -> tuple[float, float]:
    """
    Compute the signature functions of the rHEALPix map projection of
    the unit sphere.
    The north polar square is put in position `north_square`, and the
    south polar square is put in position `south_square`.

    INPUT:

    - `lam, phi` -Geographic longitude-latitude coordinates in radians.
      Assume -pi <= `lam` < pi and -pi/2 <= `phi` <= pi/2.
    - `north_square, south_square` - (Optional; defaults = 0, 0) Integers
      between 0 and 3 indicating positions of north polar and
      south polar squares, respectively.

    EXAMPLES::

        >>> print(tuple(x.tolist() for x in my_round(rhealpix_sphere(0, pi/4), 15)))
        (-1.619978633413937, 2.307012183573304)

    NOTE:

    The polar squares are labeled 0, 1, 2, 3 from east to west like this::

        east         west
        *---*---*---*---*
        | 0 | 1 | 2 | 3 |
        *---*---*---*---*
        |   |   |   |   |
        *---*---*---*---*
        | 0 | 1 | 2 | 3 |
        *---*---*---*---*
    """
    x, y = healpix_sphere(lam, phi)
    return combine_triangles(x, y, north_square=north_square, south_square=south_square)


def rhealpix_sphere_inverse(
    x: float,
    y: float,
    north_square: int = 0,
    south_square: int = 0,
    region: str = "none",
) -> tuple[float, float]:
    """
    Compute the inverse of rhealpix_sphere().

    EXAMPLES::

        >>> p = (0, pi/4)
        >>> q = rhealpix_sphere(*p)
        >>> print(tuple(x.tolist() for x in my_round(rhealpix_sphere_inverse(*q), 15)))
        (0.0, 0.785398163397448)
        >>> print(my_round(p, 15))
        (0, 0.785398163397448)

    """
    # Throw error if input coordinates are out of bounds.
    if not in_rhealpix_image(
        x, y, south_square=south_square, north_square=north_square
    ):
        print("Error (rsi): input coordinates (%.20f,%.20f) are out of bounds" % (x, y))
        return
    if region != "equatorial":
        x, y = combine_triangles(
            x, y, north_square=north_square, south_square=south_square, inverse=True
        )
    return healpix_sphere_inverse(x, y)


def rhealpix_ellipsoid(
    lam: float,
    phi: float,
    e: float = 0,
    north_square: int = 0,
    south_square: int = 0,
    region: str = "none",
) -> tuple[float, float]:
    """
    Compute the signature functions of the rHEALPix map
    projection of an oblate ellipsoid with eccentricity `e` whose
    authalic sphere is the unit sphere.
    The north polar square is put in position `north_square`,
    and the south polar square is put in position `south_square`.
    Works when `e` = 0 (spherical case) too.

    INPUT:

    - `lam, phi` - Geographic longitude-latitude coordinates in radian.
      Assume -pi <= `lam` < pi and -pi/2 <= `phi` <= pi/2.
    - `e` - Eccentricity of the ellipsoid.
    - `north_square, south_square` - (Optional; defaults = 0, 0) Integers
      between 0 and 3 indicating positions of north polar and
      south polar squares, respectively.
      See rhealpix_sphere() docstring for a diagram.

    EXAMPLES::

        >>> from numpy import arcsin
        >>> print(tuple(x if type(x) is int else x.tolist() for x in my_round(rhealpix_ellipsoid(0, arcsin(2.0/3)), 15)))
        (0, 0.785398163397448)

    """
    # Ensure north_square and south_square lie in {0, 1,2, 3}.
    x, y = healpix_ellipsoid(lam, phi, e)
    if region != "equatorial":
        x, y = combine_triangles(
            x, y, north_square=north_square, south_square=south_square
        )
    return x, y


def rhealpix_ellipsoid_inverse(
    x: float,
    y: float,
    e: float = 0,
    north_square: int = 0,
    south_square: int = 0,
    region: str = "none",
) -> tuple[float, float]:
    """
    Compute the inverse of rhealpix_ellipsoid().

    EXAMPLES::

        >>> p = (0, pi/4)
        >>> q = rhealpix_ellipsoid(*p)
        >>> print(tuple(x.tolist() for x in my_round(rhealpix_ellipsoid_inverse(*q), 15)))
        (0.0, 0.785398163397448)
        >>> print(my_round(p, 15))
        (0, 0.785398163397448)

    """
    # Throw error if input coordinates are out of bounds.
    if not in_rhealpix_image(
        x, y, south_square=south_square, north_square=north_square
    ):
        print("Error (rei): input coordinates (%.20f,%.20f) are out of bounds" % (x, y))
        return

    if region != "equatorial":
        x, y = combine_triangles(
            x, y, north_square=north_square, south_square=south_square, inverse=True
        )

    return healpix_ellipsoid_inverse(x, y, e=e)


def in_rhealpix_image(
    x: float, y: float, north_square: int = 0, south_square: int = 0
) -> bool:
    """
    Return True if and only if the point `(x, y)` lies in the image of
    the rHEALPix projection of the unit sphere.

    EXAMPLES::

        >>> eps = 0     # Test boundary points.
        >>> north_square, south_square = 0, 0
        >>> rhp = [
        ... (-pi - eps, pi/4 + eps),
        ... (-pi + north_square*pi/2 - eps, pi/4 + eps),
        ... (-pi + north_square*pi/2 - eps, 3*pi/4 + eps),
        ... (-pi + (north_square + 1)*pi/2 + eps, 3*pi/4 + eps),
        ... (-pi + (north_square + 1)*pi/2 + eps, pi/4 + eps),
        ... (pi + eps, pi/4 + eps),
        ... (pi + eps,-pi/4 - eps),
        ... (-pi + (south_square + 1)*pi/2 + eps,-pi/4 - eps),
        ... (-pi + (south_square + 1)*pi/2 + eps,-3*pi/4 - eps),
        ... (-pi + south_square*pi/2 - eps,-3*pi/4 - eps),
        ... (-pi + south_square*pi/2 -eps,-pi/4 - eps),
        ... (-pi - eps,-pi/4 - eps)
        ... ]
        >>> for p in rhp:
        ...     if not in_rhealpix_image(*p):
        ...             print('Fail')
        ...
        >>> print(in_rhealpix_image(0, 0))
        True
        >>> print(in_rhealpix_image(0, pi/4 + 0.1))
        False

    """
    # matplotlib is a third-party module.
    from matplotlib.path import Path

    # Fuzz to slightly expand rHEALPix image so that
    # points on the boundary count as lying in the image.
    eps = 1e-15
    vertices = [
        (-pi - eps, pi / 4 + eps),
        (-pi + north_square * pi / 2 - eps, pi / 4 + eps),
        (-pi + north_square * pi / 2 - eps, 3 * pi / 4 + eps),
        (-pi + (north_square + 1) * pi / 2 + eps, 3 * pi / 4 + eps),
        (-pi + (north_square + 1) * pi / 2 + eps, pi / 4 + eps),
        (pi + eps, pi / 4 + eps),
        (pi + eps, -pi / 4 - eps),
        (-pi + (south_square + 1) * pi / 2 + eps, -pi / 4 - eps),
        (-pi + (south_square + 1) * pi / 2 + eps, -3 * pi / 4 - eps),
        (-pi + south_square * pi / 2 - eps, -3 * pi / 4 - eps),
        (-pi + south_square * pi / 2 - eps, -pi / 4 - eps),
        (-pi - eps, -pi / 4 - eps),
    ]
    poly = Path(vertices)
    return bool(poly.contains_point([x, y]))


def rhealpix_vertices(
    north_square: int = 0, south_square: int = 0
) -> list[tuple[float, float]]:
    """
    Return a list of the planar vertices of the rHEALPix projection of
    the unit sphere.
    """
    vertices = [
        (pi, pi / 4),
        (-pi + (north_square + 1) * pi / 2, pi / 4),
        (-pi + (north_square + 1) * pi / 2, 3 * pi / 4),
        (-pi + north_square * pi / 2, 3 * pi / 4),
        (-pi + north_square * pi / 2, pi / 4),
        (-pi, pi / 4),
        (-pi, -pi / 4),
        (-pi + south_square * pi / 2, -pi / 4),
        (-pi + south_square * pi / 2, -3 * pi / 4),
        (-pi + (south_square + 1) * pi / 2, -3 * pi / 4),
        (-pi + (south_square + 1) * pi / 2, -pi / 4),
        (pi, -pi / 4),
    ]
    # Delete unnecessary non-vertices.
    if north_square == 3:
        vertices.remove((pi, pi / 4))
        vertices.remove((pi, pi / 4))
    elif north_square == 0:
        vertices.remove((-pi, pi / 4))
        vertices.remove((-pi, pi / 4))
    if south_square == 3:
        vertices.remove((pi, -pi / 4))
        vertices.remove((pi, -pi / 4))
    elif south_square == 0:
        vertices.remove((-pi, -pi / 4))
        vertices.remove((-pi, -pi / 4))

    return vertices


def rhealpix(
    a: float = 1,
    e: float = 0,
    north_square: int = 0,
    south_square: int = 0,
    region: str = "none",
) -> Callable[[float, float, bool, bool], tuple[float, float]]:
    """
    Return a function object that wraps the rHEALPix projection and its inverse
    of an ellipsoid with major radius `a` and eccentricity `e`.

    EXAMPLES::

        >>> f = rhealpix(a=2, e=0, north_square=1, south_square=2)
        >>> print(tuple(x.tolist() for x in my_round(f(0, pi/3, radians=True), 15)))
        (-0.574951359778215, 2.145747686573111)
        >>> p = (0, 60)
        >>> q = f(*p, radians=False)
        >>> print(tuple(x.tolist() for x in my_round(q, 15)))
        (-0.574951359778215, 2.145747686573111)
        >>> print(tuple(x.tolist() for x in my_round(f(*q, radians=False, inverse=True), 15)))
        (6e-15, 59.999999999999986)
        >>> print(my_round(p, 15))
        (0, 60)

    OUTPUT:

    - A function object of the form f(u, v, radians=False, inverse=False).
    """
    R_A = auth_rad(a, e)

    def f(
        u: float, v: float, radians: bool = False, inverse: bool = False
    ) -> tuple[float, float]:
        if not inverse:
            lam, phi = u, v
            if not radians:
                # Convert to radians.
                lam, phi = deg2rad([lam, phi])
            return tuple(
                R_A
                * array(
                    rhealpix_ellipsoid(
                        lam,
                        phi,
                        e=e,
                        north_square=north_square,
                        south_square=south_square,
                        region=region,
                    )
                )
            )
        else:
            # Scale down to R_A = 1.
            x, y = array((u, v)) / R_A
            lam, phi = array(
                rhealpix_ellipsoid_inverse(
                    x,
                    y,
                    e=e,
                    north_square=north_square,
                    south_square=south_square,
                    region=region,
                )
            )

            if not radians:
                # Convert to degrees.
                lam, phi = rad2deg([lam, phi])
            return lam, phi

    return f



import math
import csv
import argparse
import locale
from texttable import Texttable
from vgrid.generator.olcgrid import calculate_total_cells
from pyproj import Geod

geod = Geod(ellps="WGS84")

locale.setlocale(locale.LC_ALL, "")


def olc_metrics(res):
    earth_surface_area_km2 = 510_065_621.724  # 510.1 million square kilometers
    bbox = [-180, -90, 180, 90]
    num_cells = calculate_total_cells(res, bbox)
    avg_area = (earth_surface_area_km2 / num_cells) * (10**6)
    avg_edge_length = math.sqrt(avg_area)
    return num_cells, avg_edge_length, avg_area


def olc_stats(output_file=None):
    """
    Display and/or save OLC statistics to a CSV file for specific resolutions.
    """
    import locale

    locale.setlocale(locale.LC_ALL, "")  # Set locale for formatting numbers
    t = Texttable()
    t.add_row(
        ["Resolution", "Number of Cells", "Avg Edge Length (m)", "Avg Cell Area (sq m)"]
    )

    # List of specific resolutions
    resolutions = [2, 4, 6, 8, 10, 11, 12, 13, 14, 15]

    # Prepare CSV output if specified
    if output_file:
        with open(output_file, mode="w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(
                [
                    "Resolution",
                    "Number of Cells",
                    "Avg Edge Length (m)",
                    "Avg Cell Area (sq m)",
                ]
            )
            for res in resolutions:
                num_cells, avg_edge_len, avg_area = olc_metrics(res)
                writer.writerow([res, num_cells, avg_edge_len, avg_area])
    else:
        # Display in terminal
        for res in resolutions:
            num_cells, avg_edge_len, avg_area = olc_metrics(res)
            formatted_num_cells = locale.format_string("%d", num_cells, grouping=True)
            formatted_avg_edge_len = locale.format_string(
                "%.3f", avg_edge_len, grouping=True
            )
            formatted_avg_area = locale.format_string("%.3f", avg_area, grouping=True)
            t.add_row(
                [res, formatted_num_cells, formatted_avg_edge_len, formatted_avg_area]
            )
        print(t.draw())


def main():
    """
    Main function to handle command-line arguments and invoke OLC stats generation.
    """
    parser = argparse.ArgumentParser(
        description="Generate statistics for Open Location Code (OLC)."
    )
    parser.add_argument("-o", "--output", help="Output CSV file name.")
    args = parser.parse_args()

    # Generate stats
    olc_stats(args.output)


if __name__ == "__main__":
    main()



# https://github.com/ha8tks/Leaflet.Maidenhead
# https://ha8tks.github.io/Leaflet.Maidenhead/examples/
# https://www.sotamaps.org/

import json
import math
import argparse
from vgrid.utils import maidenhead
from tqdm import tqdm
from vgrid.generator.settings import max_cells, graticule_dggs_to_feature
from shapely.geometry import Polygon
from vgrid.stats.maidenheadstats import maidenhead_metrics


def generate_grid(resolution):
    if resolution == 1:
        lon_width, lat_width = 20, 10
    elif resolution == 2:
        lon_width, lat_width = 2, 1
    elif resolution == 3:
        lon_width, lat_width = 0.083333, 0.041666  # 5 minutes x 2.5 minutes
    elif resolution == 4:
        lon_width, lat_width = 0.008333, 0.004167  # 30 seconds x 15 seconds
    else:
        raise ValueError("Unsupported resolution")

    # Determine the bounding box
    min_lon, min_lat, max_lon, max_lat = [-180, -90, 180, 90]
    x_cells = int((max_lon - min_lon) / lon_width)
    y_cells = int((max_lat - min_lat) / lat_width)
    total_cells = x_cells * y_cells

    maidenhead_features = []
    with tqdm(
        total=total_cells, desc="Generating Maidenhead DGGS", unit=" cells"
    ) as pbar:
        for i in range(x_cells):
            for j in range(y_cells):
                cell_min_lon = min_lon + i * lon_width
                cell_max_lon = cell_min_lon + lon_width
                cell_min_lat = min_lat + j * lat_width
                cell_max_lat = cell_min_lat + lat_width

                cell_center_lat = (cell_min_lat + cell_max_lat) / 2
                cell_center_lat = (cell_min_lon + cell_max_lon) / 2
                maidenhead_id = maidenhead.toMaiden(
                    cell_center_lat, cell_center_lat, resolution
                )
                (
                    _,
                    _,
                    min_lat_maiden,
                    min_lon_maiden,
                    max_lat_maiden,
                    max_lon_maiden,
                    _,
                ) = maidenhead.maidenGrid(maidenhead_id)
                # Define the polygon based on the bounding box
                cell_polygon = Polygon(
                    [
                        [min_lon_maiden, min_lat_maiden],  # Bottom-left corner
                        [max_lon_maiden, min_lat_maiden],  # Bottom-right corner
                        [max_lon_maiden, max_lat_maiden],  # Top-right corner
                        [min_lon_maiden, max_lat_maiden],  # Top-left corner
                        [
                            min_lon_maiden,
                            min_lat_maiden,
                        ],  # Closing the polygon (same as the first point)
                    ]
                )
                maidenhead_feature = graticule_dggs_to_feature(
                    "maidenhead", maidenhead_id, resolution, cell_polygon
                )
                maidenhead_features.append(maidenhead_feature)
                pbar.update(1)

    return {"type": "FeatureCollection", "features": maidenhead_features}


def generate_grid_within_bbox(resolution, bbox):
    # Define the grid parameters based on the resolution
    if resolution == 1:
        lon_width, lat_width = 20, 10
    elif resolution == 2:
        lon_width, lat_width = 2, 1
    elif resolution == 3:
        lon_width, lat_width = 0.083333, 0.041666  # 5 minutes x 2.5 minutes
    elif resolution == 4:
        lon_width, lat_width = 0.008333, 0.004167  # 30 seconds x 15 seconds
    else:
        raise ValueError("Unsupported resolution")

    min_lon, min_lat, max_lon, max_lat = bbox

    # Calculate grid cell indices for the bounding box
    base_lat, base_lon = -90, -180
    start_x = math.floor((min_lon - base_lon) / lon_width)
    end_x = math.floor((max_lon - base_lon) / lon_width)
    start_y = math.floor((min_lat - base_lat) / lat_width)
    end_y = math.floor((max_lat - base_lat) / lat_width)

    maidenhead_features = []

    total_cells = (end_x - start_x + 1) * (end_y - start_y + 1)

    # Loop through all intersecting grid cells with tqdm progress bar
    with tqdm(total=total_cells, desc="Generating Maidenhead DGGS") as pbar:
        for x in range(start_x, end_x + 1):
            for y in range(start_y, end_y + 1):
                # Calculate the cell bounds
                cell_min_lon = base_lon + x * lon_width
                cell_max_lon = cell_min_lon + lon_width
                cell_min_lat = base_lat + y * lat_width
                cell_max_lat = cell_min_lat + lat_width

                # Ensure the cell intersects with the bounding box
                if not (
                    cell_max_lon < min_lon
                    or cell_min_lon > max_lon
                    or cell_max_lat < min_lat
                    or cell_min_lat > max_lat
                ):
                    # Center point for the Maidenhead code
                    cell_center_lat = (cell_min_lat + cell_max_lat) / 2
                    cell_center_lat = (cell_min_lon + cell_max_lon) / 2

                    maidenhead_id = maidenhead.toMaiden(
                        cell_center_lat, cell_center_lat, resolution
                    )
                    (
                        _,
                        _,
                        min_lat_maiden,
                        min_lon_maiden,
                        max_lat_maiden,
                        max_lon_maiden,
                        _,
                    ) = maidenhead.maidenGrid(maidenhead_id)
                    # Define the polygon based on the bounding box
                    cell_polygon = Polygon(
                        [
                            [min_lon_maiden, min_lat_maiden],  # Bottom-left corner
                            [max_lon_maiden, min_lat_maiden],  # Bottom-right corner
                            [max_lon_maiden, max_lat_maiden],  # Top-right corner
                            [min_lon_maiden, max_lat_maiden],  # Top-left corner
                            [
                                min_lon_maiden,
                                min_lat_maiden,
                            ],  # Closing the polygon (same as the first point)
                        ]
                    )

                    maidenhead_feature = graticule_dggs_to_feature(
                        "maidenhead", maidenhead_id, resolution, cell_polygon
                    )

                    maidenhead_features.append(maidenhead_feature)

                pbar.update(1)

    return {"type": "FeatureCollection", "features": maidenhead_features}


def main():
    parser = argparse.ArgumentParser(description="Generate Maidenhead DGGS")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=[1, 2, 3, 4],
        default=1,
        help="resolution [1..4]",
    )
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )
    args = parser.parse_args()
    resolution = args.resolution
    bbox = args.bbox if args.bbox else [-180, -90, 180, 90]

    if bbox == [-180, -90, 180, 90]:
        # Calculate the number of cells at the given resolution
        num_cells, _, _ = maidenhead_metrics(resolution)
        print(f"Resolution {resolution} will generate {num_cells} cells ")
        if num_cells > max_cells:
            print(f"which exceeds the limit of {max_cells}.")
            print("Please select a smaller resolution and try again.")
            return
        geojson_features = generate_grid(resolution)
    else:
        geojson_features = generate_grid_within_bbox(resolution, bbox)

    if geojson_features:
        # Define the GeoJSON file path
        geojson_path = f"maidenhead_grid_{resolution}.geojson"
        with open(geojson_path, "w") as f:
            json.dump(geojson_features, f, indent=2)

        print(f"GeoJSON saved as {geojson_path}")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 15 14:57:52 2018

@author: Joseph
"""

"""
This Python 3.3 module implements the icosohedral Snyder equal area map projection (ISEA) for ellipsoids of revolution by using the spherical version of the projection (from the PROJ.4 library) on the ellipsoid's authalic sphere.
CHANGELOG:
- Alexander Raichev (AR), 2012-01-25: Initial version.
- AR, 2013-07-23: Ported to Python 3.3.
NOTE:
All lengths are measured in meters and all angles are measured in radians 
unless indicated otherwise. 
By 'ellipsoid' below, i mean an oblate ellipsoid of revolution.
"""
# *****************************************************************************
#       Copyright (C) 2012 Alexander Raichev <alex.raichev@gmail.com>
#
#  Distributed under the terms of the GNU Lesser General Public License (LGPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************

# Import third-party modules.
import pyproj
from numpy import rad2deg, deg2rad, array

# Import my modules.
from .utils import auth_lat, auth_rad


def isea_sphere(lam, phi):
    r"""
    Compute the signature function of the icosohedral Snyder equal area
    (ISEA) map projection of the unit sphere.

    INPUT:

    - `lam, phi` - Geodetic longitude-latitude coordinates in radians.
      Assume -pi <= `lam` < pi and -pi/2 <= `phi` <= pi/2.

    EXAMPLES::

    """
    f = pyproj.Proj(proj="isea", R=1)
    return f(lam, phi, radians=True)


def isea_sphere_inverse(x, y):
    r"""
    Compute the inverse of isea_sphere().
    INPUT:

    - `x, y` - Planar coordinates in meters in the image of the
      ISEA projection of the unit sphere.

    NOTES:

    Not implemented here nor in PROJ.4.
    """
    pass


def isea_ellipsoid(lam, phi, e=0):
    r"""
    Compute the signature functions of the ISEA may projection of an oblate
    ellipsoid with eccentricity `e` whose authalic sphere is the unit sphere.

    INPUT:

    - `lam, phi` - Geodetic longitude-latitude coordinates in radians.
      Assume -pi <= `lam` < pi and -pi/2 <= `phi` <= pi/2.
    - `e` - Eccentricity of the oblate ellipsoid.
    """
    beta = auth_lat(phi, e, radians=True)
    return isea_sphere(lam, beta)


def isea_ellipsoid_inverse(x, y, e=0):
    r"""
    Compute the inverse of isea_ellipsoid().

    INPUT:

    - `x, y` - Planar coordinates in meters in the image of the
      ISEA projection of the ellipsoid with eccentricity `e` whose
      authalic sphere is the unit sphere.

    NOTES:

    Not implemented here nor in PROJ.4.
    """
    pass


def isea(a=1, e=0):
    r"""
    Return a function object that wraps the ISEA projection and its inverse
    of an ellipsoid with major radius `a` and eccentricity `e`.

    EXAMPLES::

        >>> f = isea(a=2, e=0)
        >>> print(my_round(f(0, pi/3, radians=True), 15))
        (-4.7097959155097699, 2.9195761776404252)
        >>> g = isea(a=2, e=0.1)
        >>> print(my_round(g(0, 60, radians=False), 15))
        (-4.6978889550868486, 2.9179977608222689)

    OUTPUT:

    - A function object of the form f(u, v, radians=False, inverse=False).
    """
    R_A = auth_rad(a, e)

    def f(u, v, radians=False, inverse=False):
        if not inverse:
            lam, phi = u, v
            if not radians:
                # Convert to radians.
                lam, phi = deg2rad([lam, phi])
            return tuple(R_A * array(isea_ellipsoid(lam, phi, e=e)))
        else:
            # Scale down to R_A = 1.
            x, y = array((u, v)) / R_A
            lam, phi = array(isea_ellipsoid_inverse(x, y, e=e))
            if not radians:
                # Convert to degrees.
                lam, phi = rad2deg([lam, phi])
            return lam, phi

    return f



import numpy as np
from shapely.geometry import shape, Polygon
from shapely.ops import transform
from pyproj import CRS, Transformer
import argparse
import re
from tqdm import tqdm
from vgrid.generator.settings import graticule_dggs_to_feature
from vgrid.utils import mgrs
import json
import os


def is_valid_gzd(gzd):
    """Check if a Grid Zone Designator (GZD) is valid."""
    pattern = r"^(?:0[1-9]|[1-5][0-9]|60)[C-HJ-NP-X]$"
    return bool(re.match(pattern, gzd))


def generate_grid(gzd, resolution):  # Define the UTM CRS
    # Reference: https://www.maptools.com/tutorials/utm/details
    cell_size = 100_000 // (10**resolution)
    north_bands = "NPQRSTUVWX"
    south_bands = "MLKJHGFEDC"
    band_distance = 111_132 * 8
    gzd_band = gzd[2]

    if gzd_band >= "N":  # North Hemesphere
        epsg_code = int("326" + gzd[:2])
        min_x, min_y, max_x, max_y = 100000, 0, 900000, 9500000  # for the North
        north_band_idx = north_bands.index(gzd_band)
        max_y = band_distance * (north_band_idx + 1)
        if gzd_band == "X":
            max_y += band_distance  # band X = 12 deggrees instead of 8 degrees

    else:  # South Hemesphere
        epsg_code = int("327" + gzd[:2])
        min_x, min_y, max_x, max_y = 100000, 0, 900000, 10000000  # for the South
        south_band_idx = south_bands.index(gzd_band)
        max_y = band_distance * (south_band_idx + 1)

    utm_crs = CRS.from_epsg(epsg_code)
    wgs84_crs = CRS.from_epsg(4326)
    transformer = Transformer.from_crs(utm_crs, wgs84_crs, always_xy=True).transform

    gzd_json_path = os.path.join(os.path.dirname(__file__), "gzd.geojson")
    with open(gzd_json_path, "r") as f:
        gzd_data = json.load(f)

    gzd_features = gzd_data["features"]
    gzd_feature = [
        feature for feature in gzd_features if feature["properties"].get("gzd") == gzd
    ][0]
    gzd_geom = shape(gzd_feature["geometry"])

    # Create grid polygons
    mgrs_features = []
    x_coords = np.arange(min_x, max_x, cell_size)
    y_coords = np.arange(min_y, max_y, cell_size)
    num_cells = len(x_coords) * len(y_coords)
    with tqdm(total=num_cells, desc="Generating MGRS DGGS", unit=" cells") as pbar:
        for x in x_coords:
            for y in y_coords:
                cell_polygon_utm = Polygon(
                    [
                        (x, y),
                        (x + cell_size, y),
                        (x + cell_size, y + cell_size),
                        (x, y + cell_size),
                        (x, y),  # Close the polygon
                    ]
                )
                cell_polygon = transform(transformer, cell_polygon_utm)

                if cell_polygon.intersects(gzd_geom):
                    centroid_lat, centroid_lon = (
                        cell_polygon.centroid.y,
                        cell_polygon.centroid.x,
                    )
                    mgrs_id = mgrs.toMgrs(centroid_lat, centroid_lon, resolution)
                    mgrs_feature = graticule_dggs_to_feature(
                        "mgrs", mgrs_id, resolution, cell_polygon
                    )
                    # clip inside GZD:
                    if not gzd_geom.contains(cell_polygon):
                        intersected_polygon = cell_polygon.intersection(gzd_geom)
                        if intersected_polygon:
                            intersected_centroid_lat, intersected_centroid_lon = (
                                intersected_polygon.centroid.y,
                                intersected_polygon.centroid.x,
                            )
                            interescted_mgrs_id = mgrs.toMgrs(
                                intersected_centroid_lat,
                                intersected_centroid_lon,
                                resolution,
                            )
                            mgrs_feature = graticule_dggs_to_feature(
                                "mgrs",
                                interescted_mgrs_id,
                                resolution,
                                intersected_polygon,
                            )
                    mgrs_features.append(mgrs_feature)
                pbar.update(1)
    return {"type": "FeatureCollection", "features": mgrs_features}


def main():
    # Set up argument parser
    parser = argparse.ArgumentParser(description="Generate MGRS DGGS.")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        default=0,
        required=True,
        help="Resolution [0..5]",
    )
    parser.add_argument(
        "-gzd",
        type=str,
        default="48P",
        required=True,
        help="GZD - Grid Zone Designator, e.g. 48P",
    )
    # Parse the arguments
    args = parser.parse_args()

    gzd = args.gzd
    if not is_valid_gzd(gzd):
        print("Invalid GZD. Please input a valid GZD and try again.")
        return

    resolution = args.resolution
    if resolution < 0 or resolution > 5:
        print("Please select a resolution in [0..5] range and try again ")
        return

    geojson_features = generate_grid(gzd, resolution)

    if geojson_features:
        geojson_path = f"mgrs_grid_{gzd}_{resolution}.geojson"
        with open(geojson_path, "w") as f:
            json.dump(geojson_features, f, indent=2)

        print(f"GeoJSON saved as {geojson_path}")


if __name__ == "__main__":
    main()



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

from vgrid.utils.eaggr.shapes.dggs_linestring import DggsLinestring


## Defines a DGGS polygon made up of an outer ring and zero or more inner rings.
#  The outer and inner rings are linestrings made up of DGGS cells.
class DggsPolygon(object):
    ## Sets the outer ring of the polygon. Initially the polygon has no inner rings.
    #  @note Call add_inner_ring() to add inner rings to the polygon.
    #  @param outer_ring DggsLinestring defining the outer ring of the polygon.
    #  @throw ValueError Thrown if input argument is not a DggsLinestring object.
    def __init__(self, outer_ring):
        if not isinstance(outer_ring, DggsLinestring):
            raise ValueError("Input argument to DggsPolygon() must be a DggsLinestring")
        self._outer_ring = outer_ring
        self._inner_rings = []

    ## Adds an inner ring to the polygon.
    #  @param inner_ring DggsLinestring defining a new inner ring for the polygon.
    #  @throw ValueError Thrown if input argument is not a DggsLinestring object.
    def add_inner_ring(self, inner_ring):
        if not isinstance(inner_ring, DggsLinestring):
            raise ValueError(
                "Input argument to add_inner_ring() must be a DggsLinestring"
            )
        self._inner_rings.append(inner_ring)

    ## @return DggsLinestring defining the outer ring of the polygon.
    def get_outer_ring(self):
        return self._outer_ring

    ## @return List of DggsLinestring objects which define the inner rings of the polygon.
    def get_inner_rings(self):
        return self._inner_rings



"""
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""


def maidenGridCenter(maiden):
    """
    maiden is a string the is even and between 2 to 8 characters
    It returns the center point of a maiden grid and throws an exception
    on error
    """
    if not isinstance(maiden, str):
        raise TypeError("Maidenhead locator must be a string")

    maiden = maiden.strip().upper()

    N = len(maiden)
    if not 8 >= N >= 2 and N % 2 == 0:
        raise ValueError(
            "Maidenhead locator requires 2-8 characters, even number of characters"
        )

    Oa = ord("A")
    lon = -180.0
    lat = -90.0
    # %% first pair
    isValid(maiden[0], 0)
    isValid(maiden[1], 0)
    lon += (ord(maiden[0]) - Oa) * 20
    lat += (ord(maiden[1]) - Oa) * 10
    if N == 2:
        lon += 10
        lat += 5
    # %% second pair
    if N >= 4:
        isValid(maiden[2], 1)
        isValid(maiden[3], 1)
        lon += int(maiden[2]) * 2
        lat += int(maiden[3]) * 1
    if N == 4:
        lon += 1
        lat += 0.5
    # %%
    if N >= 6:
        isValid(maiden[4], 2)
        isValid(maiden[5], 2)
        lon += (ord(maiden[4]) - Oa) * 5.0 / 60
        lat += (ord(maiden[5]) - Oa) * 2.5 / 60
    if N == 6:
        lon += 5.0 / 120
        lat += 2.5 / 120
    # %%
    if N == 8:
        lon += int(maiden[6]) * 5.0 / 600
        lat += int(maiden[7]) * 2.5 / 600
        lon += 5.0 / 1200
        lat += 2.5 / 1200

    return lat, lon


def maidenGrid(maiden):
    """
    maiden is a string the is even and between 2 to 8 characters
    It returns the center point of a maiden grid and throws an exception
    on error
    """
    if not isinstance(maiden, str):
        raise TypeError("Maidenhead locator must be a string")

    maiden = maiden.strip().upper()

    N = len(maiden)
    if not 8 >= N >= 2 and N % 2 == 0:
        raise ValueError(
            "Maidenhead locator requires 2-8 characters, even number of characters"
        )

    Oa = ord("A")
    lon = -180.0
    lat = -90.0
    # %% first pair
    isValid(maiden[0], 0)
    isValid(maiden[1], 0)
    lon += (ord(maiden[0]) - Oa) * 20
    lat += (ord(maiden[1]) - Oa) * 10
    if N == 2:
        lon1 = lon
        lat1 = lat
        lon2 = lon + 20
        lat2 = lat + 10
        lon += 10
        lat += 5
    # %% second pair
    if N >= 4:
        isValid(maiden[2], 1)
        isValid(maiden[3], 1)
        lon += int(maiden[2]) * 2
        lat += int(maiden[3]) * 1
    if N == 4:
        lon1 = lon
        lat1 = lat
        lon2 = lon + 2
        lat2 = lat + 1
        lon += 1
        lat += 0.5
    # %%
    if N >= 6:
        isValid(maiden[4], 2)
        isValid(maiden[5], 2)
        lon += (ord(maiden[4]) - Oa) * 5.0 / 60
        lat += (ord(maiden[5]) - Oa) * 2.5 / 60
    if N == 6:
        lon1 = lon
        lat1 = lat
        lon2 = lon + 5.0 / 60
        lat2 = lat + 2.5 / 60
        lon += 5.0 / 120
        lat += 2.5 / 120
    # %%
    if N == 8:
        lon += int(maiden[6]) * 5.0 / 600
        lat += int(maiden[7]) * 2.5 / 600
        lon1 = lon
        lat1 = lat
        lon2 = lon + 5.0 / 600
        lat2 = lat + 2.5 / 600
        lon += 5.0 / 1200
        lat += 2.5 / 1200

    return lat, lon, lat1, lon1, lat2, lon2, maiden


def isValid(c, level):
    if level == 0:
        if not "R" >= c >= "A":
            raise ValueError("Invalid maidenhead encoding")
    if level == 1 or level == 3:
        if not "9" >= c >= "0":
            raise ValueError("Invalid maidenhead encoding")
    if level == 2:
        if not "X" >= c >= "A":
            raise ValueError("Invalid maidenhead encoding")
    return True


def toMaiden(lat, lon=None, precision=3):
    """
    Returns a maidenhead string for latitude, longitude at specified level.

    Parameters
    ----------

    lat : float or tuple of float
        latitude or tuple of latitude, longitude
    lon : float, optional
        longitude (if not given tuple)
    precision : int, optional
        level of precision (length of maidenhead grid string output)

    Returns
    -------

    maiden : str
        Maidenhead grid string of specified precision
    """

    if lon < -180.0 or lon > 180.0 or lat < -90.0 or lat > 90.0:
        raise ValueError("Maidenhead: invalid latitude and longitude")
    A = ord("A")
    a = divmod(lon + 180, 20)
    b = divmod(lat + 90, 10)
    maiden = chr(A + int(a[0])) + chr(A + int(b[0]))
    lon = a[1] / 2.0
    lat = b[1]
    i = 1
    while i < precision:
        i += 1
        a = divmod(lon, 1)
        b = divmod(lat, 1)
        if not (i % 2):
            maiden += str(int(a[0])) + str(int(b[0]))
            lon = 24 * a[1]
            lat = 24 * b[1]
        else:
            maiden += chr(A + int(a[0])) + chr(A + int(b[0]))
            lon = 10 * a[1]
            lat = 10 * b[1]

    if len(maiden) >= 6:
        maiden = maiden[:4] + maiden[4:6].lower() + maiden[6:]

    return maiden



import json
import argparse
from tqdm import tqdm
from shapely.geometry import Polygon, box
import numpy as np
from vgrid.utils.gars.garsgrid import GARSGrid  # Ensure the correct import path

from vgrid.generator.settings import max_cells, graticule_dggs_to_feature


def get_resolution_minutes(resolution):
    """Convert resolution level to minutes.

    Parameters
    ----------
    resolution : int
        Resolution level (1-4):
        1 = 30 minutes (minimum resolution)
        2 = 15 minutes
        3 = 5 minutes
        4 = 1 minute (maximum resolution)

    Returns
    -------
    int
        Resolution in minutes
    """
    minutes_map = {
        1: 30,  # 30 minutes
        2: 15,  # 15 minutes
        3: 5,  # 5 minutes
        4: 1,  # 1 minute
    }
    return minutes_map[resolution]


def generate_grid(resolution):
    # Default to the whole world if no bounding box is provided
    lon_min, lat_min, lon_max, lat_max = -180, -90, 180, 90

    resolution_minutes = get_resolution_minutes(resolution)
    resolution_degrees = resolution_minutes / 60.0

    # Generate ranges for longitudes and latitudes
    longitudes = np.arange(lon_min, lon_max, resolution_degrees)
    latitudes = np.arange(lat_min, lat_max, resolution_degrees)

    total_cells = len(longitudes) * len(latitudes)

    gars_features = []
    # Loop over longitudes and latitudes with tqdm progress bar
    with tqdm(total=total_cells, desc="Generating GARS DGGS", unit=" cells") as pbar:
        for lon in longitudes:
            for lat in latitudes:
                # Create the GARS grid code
                gars_cell = GARSGrid.from_latlon(lat, lon, resolution_minutes)
                wkt_polygon = gars_cell.polygon

                if wkt_polygon:
                    cell_polygon = Polygon(list(wkt_polygon.exterior.coords))
                    gars_id = gars_cell.gars_id
                    gars_feature = graticule_dggs_to_feature(
                        "gars", gars_id, resolution, cell_polygon
                    )
                    gars_features.append(gars_feature)
                    pbar.update(1)

    # Create a FeatureCollection
    return {
        "type": "FeatureCollection",
        "features": gars_features,
    }


def generate_grid_within_bbox(bbox, resolution):
    # Default to the whole world if no bounding box is provided
    bbox_polygon = box(*bbox)
    lon_min, lat_min, lon_max, lat_max = bbox

    resolution_minutes = get_resolution_minutes(resolution)
    resolution_degrees = resolution_minutes / 60.0

    longitudes = np.arange(
        lon_min - resolution_degrees, lon_max + resolution_degrees, resolution_degrees
    )
    latitudes = np.arange(
        lat_min - resolution_degrees, lat_max + resolution_degrees, resolution_degrees
    )

    gars_features = []
    # Loop over longitudes and latitudes with tqdm progress bar
    with tqdm(desc="Generating GARS DGGS", unit=" cells") as pbar:
        for lon in longitudes:
            for lat in latitudes:
                # Create the GARS grid code
                gars_cell = GARSGrid.from_latlon(lat, lon, resolution_minutes)
                wkt_polygon = gars_cell.polygon

                if wkt_polygon:
                    cell_polygon = Polygon(list(wkt_polygon.exterior.coords))

                    if bbox_polygon.intersects(cell_polygon):
                        gars_id = gars_cell.gars_id
                        gars_feature = graticule_dggs_to_feature(
                            "gars", gars_id, resolution, cell_polygon
                        )
                        gars_features.append(gars_feature)
                        pbar.update(1)

    # Create a FeatureCollection
    return {
        "type": "FeatureCollection",
        "features": gars_features,
    }


def main():
    parser = argparse.ArgumentParser(description="Generate GARS DGGS")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=[1, 2, 3, 4],
        required=True,
        help="Resolution level (1=30min, 2=15min, 3=5min, 4=1min)",
    )
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )

    args = parser.parse_args()
    resolution = args.resolution
    bbox = args.bbox if args.bbox else [-180, -90, 180, 90]

    # Write to a GeoJSON file
    if bbox == [-180, -90, 180, 90]:
        # Calculate the number of cells at the given resolution
        lon_min, lat_min, lon_max, lat_max = -180, -90, 180, 90

        resolution_minutes = get_resolution_minutes(resolution)
        resolution_degrees = resolution_minutes / 60.0

        # Generate ranges for longitudes and latitudes
        longitudes = np.arange(lon_min, lon_max, resolution_degrees)
        latitudes = np.arange(lat_min, lat_max, resolution_degrees)

        total_cells = len(longitudes) * len(latitudes)
        print(
            f"Resolution level {resolution} ({resolution_minutes} minutes) will generate {total_cells} cells "
        )
        if total_cells > max_cells:
            print(f"which exceeds the limit of {max_cells}.")
            print("Please select a smaller resolution and try again.")
            return

        geojson_features = generate_grid(resolution)

    else:
        geojson_features = generate_grid_within_bbox(bbox, resolution)

    output_filename = f"gars_grid_{resolution}.geojson"

    with open(output_filename, "w") as f:
        json.dump(geojson_features, f, indent=2)

    print(f"GARS grid saved to {output_filename}")


if __name__ == "__main__":
    main()



import argparse
import json
from shapely.geometry import Polygon
import platform

if platform.system() == "Windows":
    from vgrid.utils.eaggr.eaggr import Eaggr
    from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
    from vgrid.utils.eaggr.enums.model import Model
    from vgrid.utils.eaggr.enums.shape_string_format import ShapeStringFormat

from tqdm import tqdm
from shapely.geometry import box, mapping
from vgrid.utils.antimeridian import fix_polygon
import platform
from vgrid.generator.settings import (
    max_cells,
    isea3h_base_cells,
    isea3h_accuracy_res_dict,
    isea3h_res_accuracy_dict,
)

from pyproj import Geod

geod = Geod(ellps="WGS84")


def isea3h_cell_to_polygon(isea3h_dggs, isea3h_cell):
    if platform.system() == "Windows":
        cell_to_shape = isea3h_dggs.convert_dggs_cell_outline_to_shape_string(
            isea3h_cell, ShapeStringFormat.WKT
        )
        if cell_to_shape:
            coordinates_part = cell_to_shape.replace("POLYGON ((", "").replace("))", "")
            coordinates = []
            for coord_pair in coordinates_part.split(","):
                lon, lat = map(float, coord_pair.strip().split())
                coordinates.append([lon, lat])

            # Ensure the polygon is closed (first and last point must be the same)
            if coordinates[0] != coordinates[-1]:
                coordinates.append(coordinates[0])

        cell_polygon = Polygon(coordinates)
        fixed_polygon = fix_polygon(cell_polygon)
        return fixed_polygon


def get_isea3h_children_cells(isea3h_dggs, base_cells, target_resolution):
    """
    Recursively generate DGGS cells for the desired resolution, avoiding duplicates.
    """
    if platform.system() == "Windows":
        current_cells = base_cells
        seen_cells = set(base_cells)  # Track already processed cells

        for res in range(target_resolution):
            next_cells = []
            for cell in current_cells:
                children = isea3h_dggs.get_dggs_cell_children(DggsCell(cell))
                for child in children:
                    if child._cell_id not in seen_cells:
                        seen_cells.add(child._cell_id)  # Mark as seen
                        next_cells.append(child._cell_id)
            current_cells = next_cells
        return current_cells


def get_isea3h_children_cells_within_bbox(
    isea3h_dggs, bounding_cell, bbox, target_resolution
):
    """
    Recursively generate DGGS cells within a bounding box, avoiding duplicates.
    """
    if platform.system() == "Windows":
        current_cells = [
            bounding_cell
        ]  # Start with a list containing the single bounding cell
        seen_cells = set(current_cells)  # Track already processed cells
        bounding_cell2point = isea3h_dggs.convert_dggs_cell_to_point(
            DggsCell(bounding_cell)
        )
        accuracy = bounding_cell2point._accuracy
        bounding_resolution = isea3h_accuracy_res_dict.get(accuracy)

        if bounding_resolution <= target_resolution:
            for res in range(bounding_resolution, target_resolution):
                next_cells = []
                for cell in current_cells:
                    # Get the child cells for the current cell
                    children = isea3h_dggs.get_dggs_cell_children(DggsCell(cell))
                    for child in children:
                        if (
                            child._cell_id not in seen_cells
                        ):  # Check if the child is already processed
                            child_shape = isea3h_cell_to_polygon(isea3h_dggs, child)
                            if child_shape.intersects(bbox):
                                seen_cells.add(child._cell_id)  # Mark as seen
                                next_cells.append(child._cell_id)
                if not next_cells:  # Break early if no cells remain
                    break
                current_cells = next_cells  # Update current_cells to process the next level of children

            return current_cells
        else:
            # print('Bounding box area is < 0.028 square meters. Please select a bigger bounding box')
            return None


def generate_grid(isea3h_dggs, resolution):
    """
    Generate DGGS cells and convert them to GeoJSON features.
    """
    if platform.system() == "Windows":
        children = get_isea3h_children_cells(isea3h_dggs, isea3h_base_cells, resolution)
        features = []
        for child in tqdm(children, desc="Generating ISEA3H DGGS", unit=" cells"):
            isea3h_cell = DggsCell(child)
            cell_polygon = isea3h_cell_to_polygon(isea3h_dggs, isea3h_cell)
            isea3h_id = isea3h_cell.get_cell_id()

            cell_centroid = cell_polygon.centroid
            center_lat = round(cell_centroid.y, 7)
            center_lon = round(cell_centroid.x, 7)
            cell_area = round(abs(geod.geometry_area_perimeter(cell_polygon)[0]), 3)
            cell_perimeter = abs(geod.geometry_area_perimeter(cell_polygon)[1])
            avg_edge_len = round(cell_perimeter / 6, 3)
            if resolution == 0:
                avg_edge_len = round(cell_perimeter / 3, 3)  # icosahedron faces

            features.append(
                {
                    "type": "Feature",
                    "geometry": mapping(cell_polygon),
                    "properties": {
                        "isea3h": isea3h_id,
                        "resolution": resolution,
                        "center_lat": center_lat,
                        "center_lon": center_lon,
                        "cell_area": cell_area,
                        "avg_edge_len": avg_edge_len,
                    },
                }
            )

        return {"type": "FeatureCollection", "features": features}


def generate_grid_within_bbox(isea3h_dggs, resolution, bbox):
    if platform.system() == "Windows":
        accuracy = isea3h_res_accuracy_dict.get(resolution)
        # print(accuracy)
        bounding_box = box(*bbox)
        bounding_box_wkt = bounding_box.wkt  # Create a bounding box polygon
        # print (bounding_box_wkt)
        shapes = isea3h_dggs.convert_shape_string_to_dggs_shapes(
            bounding_box_wkt, ShapeStringFormat.WKT, accuracy
        )
        shape = shapes[0]
        # for shape in shapes:
        bbox_cells = shape.get_shape().get_outer_ring().get_cells()
        bounding_cell = isea3h_dggs.get_bounding_dggs_cell(bbox_cells)
        # print("boudingcell: ", bounding_cell.get_cell_id())
        bounding_children_cells = get_isea3h_children_cells_within_bbox(
            isea3h_dggs, bounding_cell.get_cell_id(), bounding_box, resolution
        )
        # print (bounding_children_cells)
        if bounding_children_cells:
            features = []
            for child in bounding_children_cells:
                isea3h_cell = DggsCell(child)
                cell_polygon = isea3h_cell_to_polygon(isea3h_dggs, isea3h_cell)
                isea3h_id = isea3h_cell.get_cell_id()

                cell_centroid = cell_polygon.centroid
                center_lat = round(cell_centroid.y, 7)
                center_lon = round(cell_centroid.x, 7)
                cell_area = round(abs(geod.geometry_area_perimeter(cell_polygon)[0]), 3)
                cell_perimeter = abs(geod.geometry_area_perimeter(cell_polygon)[1])
                avg_edge_len = round(cell_perimeter / 6, 3)
                if resolution == 0:
                    avg_edge_len = round(cell_perimeter / 3, 3)  # icosahedron faces

                # if cell_polygon.intersects(bounding_box):
                features.append(
                    {
                        "type": "Feature",
                        "geometry": mapping(cell_polygon),
                        "properties": {
                            "isea3h": isea3h_id,
                            "resolution": resolution,
                            "center_lat": center_lat,
                            "center_lon": center_lon,
                            "cell_area": cell_area,
                            "avg_edge_len": avg_edge_len,
                        },
                    }
                )

            return {"type": "FeatureCollection", "features": features}


def main():
    parser = argparse.ArgumentParser(description="Generate Open-Eaggr ISEA3H DGGS.")
    parser.add_argument(
        "-r", "--resolution", type=int, required=True, help="Resolution [0..32]"
    )
    # Resolution max range: [0..40]
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )
    if platform.system() == "Windows":
        isea3h_dggs = Eaggr(Model.ISEA3H)
        args = parser.parse_args()
        resolution = args.resolution
        bbox = args.bbox if args.bbox else [-180, -90, 180, 90]
        if resolution < 0 or resolution > 32:
            print("Please select a resolution in [0..32] range and try again ")
            return

        if bbox == [-180, -90, 180, 90]:
            total_cells = 20 * (7**resolution)
            print(
                f"Resolution {resolution} within bounding box {bbox} will generate {total_cells} cells "
            )

            if total_cells > max_cells:
                print(f"which exceeds the limit of {max_cells}. ")
                print("Please select a smaller resolution and try again.")
                return

            geojson_features = generate_grid(isea3h_dggs, resolution)

        else:
            # Generate grid within the bounding box
            geojson_features = generate_grid_within_bbox(isea3h_dggs, resolution, bbox)

        # Define the GeoJSON file path
        geojson_path = f"isea3h_grid_{resolution}.geojson"
        with open(geojson_path, "w") as f:
            json.dump(geojson_features, f, indent=2)

        print(f"GeoJSON saved as {geojson_path}")


if __name__ == "__main__":
    main()



import json
import os
import argparse


def fix_antimeridian_cells(hex_boundary, threshold=-128):
    if any(lon < threshold for _, lon in hex_boundary):
        # Adjust all longitudes accordingly
        return [(lat, lon - 360 if lon > 0 else lon) for lat, lon in hex_boundary]
    return hex_boundary


def fix_geojson(geojson_data, threshold=-128):
    """
    Apply the fix_antimeridian_cells function to all geometries in the GeoJSON data.
    """
    for feature in geojson_data.get("features", []):
        geometry = feature.get("geometry", {})
        if geometry["type"] == "Polygon":
            # Fix the exterior ring
            geometry["coordinates"][0] = fix_antimeridian_cells(
                geometry["coordinates"][0], threshold
            )
            # Fix any interior rings (holes)
            for i in range(1, len(geometry["coordinates"])):
                geometry["coordinates"][i] = fix_antimeridian_cells(
                    geometry["coordinates"][i], threshold
                )
        elif geometry["type"] == "MultiPolygon":
            for polygon in geometry["coordinates"]:
                # Fix the exterior ring
                polygon[0] = fix_antimeridian_cells(polygon[0], threshold)
                # Fix any interior rings (holes)
                for i in range(1, len(polygon)):
                    polygon[i] = fix_antimeridian_cells(polygon[i], threshold)
    return geojson_data


def fix_geom(input_file, output_file):
    with open(input_file, "r") as infile:
        geojson_data = json.load(infile)  # Load as dictionary

    fixed_geojson_data = fix_geojson(geojson_data)
    with open(output_file, "w") as outfile:
        json.dump(fixed_geojson_data, outfile, indent=2)

    print(f"GeoJSON has been fixed for the antimeridian and saved to {output_file}")


def main():
    # Set up argument parser
    parser = argparse.ArgumentParser(
        description="Fix antimeridian GeoJSON created from DGGRID"
    )
    parser.add_argument("input", help="Input GeoJSON file")

    # Parse the arguments
    args = parser.parse_args()

    # Derive output filename by appending "_fixed" to the input filename
    output_file = os.path.join(
        os.getcwd(),
        f"{os.path.splitext(os.path.basename(args.input))[0]}_geom_fixed_2.geojson",
    )

    # Call the function with the input and output file
    fix_geom(args.input, output_file)


if __name__ == "__main__":
    main()



from vgrid.utils import s2, olc, mercantile
import h3
import argparse
import json
from tqdm import tqdm
import os
import re
from vgrid.stats.s2stats import s2_metrics
from vgrid.stats.rhealpixstats import rhealpix_metrics
from vgrid.stats.isea4tstats import isea4t_metrics
from vgrid.stats.qtmstats import qtm_metrics
from vgrid.stats.olcstats import olc_metrics
from vgrid.stats.geohashstats import geohash_metrics
from vgrid.stats.tilecodestats import tilecode_metrics
from vgrid.stats.quadkeystats import quadkey_metrics

from shapely.geometry import shape
from vgrid.generator import (
    h3grid,
    s2grid,
    rhealpixgrid,
    isea4tgrid,
    qtmgrid,
    olcgrid,
    geohashgrid,
    tilecodegrid,
    quadkeygrid,
)
from numbers import Number
from vgrid.utils.rhealpixdggs.dggs import RHEALPixDGGS
from vgrid.utils.rhealpixdggs.ellipsoids import WGS84_ELLIPSOID
from pyproj import Geod
import platform

if platform.system() == "Windows":
    from vgrid.utils.eaggr.eaggr import Eaggr
    from vgrid.utils.eaggr.enums.model import Model


geod = Geod(ellps="WGS84")
E = WGS84_ELLIPSOID


def get_nearest_resolution(geojson_features, from_dggs, to_dggs, from_field=None):
    if not (from_field):
        from_field = from_dggs
    try:
        from_dggs_id = geojson_features["features"][0]["properties"][from_field]
    except Exception:
        print(f"There is no valid DGGS IDs found in <{from_field}> field.")
        return
    try:
        if from_dggs == "h3":
            from_resolution = h3.get_resolution(from_dggs_id)
            from_area = h3.average_hexagon_area(from_resolution, unit="m^2")

        elif from_dggs == "s2":
            s2_id = s2.CellId.from_token(from_dggs_id)
            from_resolution = s2_id.level()
            _, _, from_area = s2_metrics(from_resolution)

        elif from_dggs == "rhealpix":
            rhealpix_uids = (from_field[0],) + tuple(map(int, from_field[1:]))
            rhealpix_dggs = RHEALPixDGGS(
                ellipsoid=E, north_square=1, south_square=3, N_side=3
            )
            rhealpix_cell = rhealpix_dggs.cell(rhealpix_uids)
            from_resolution = rhealpix_cell.resolution
            _, _, from_area = rhealpix_metrics(from_resolution)

        elif from_dggs == "isea4t":
            if platform.system() == "Windows":
                isea4t_dggs = Eaggr(Model.ISEA4T)
                from_resolution = len(from_field) - 2
                _, _, from_area, _ = isea4t_metrics(isea4t_dggs, from_resolution)

        elif from_dggs == "qtm":
            from_resolution = len(from_field)
            _, _, from_area = qtm_metrics(from_resolution)

        elif from_dggs == "olc":
            coord = olc.decode(from_field)
            from_resolution = coord.codeLength
            _, _, from_area = olc_metrics(from_resolution)

        elif from_dggs == "geohash":
            from_resolution = len(from_field)
            _, _, from_area = geohash_metrics(from_resolution)

        elif from_dggs == "tilecode":
            match = re.match(r"z(\d+)x(\d+)y(\d+)", from_field)
            from_resolution = int(match.group(1))
            _, _, from_area = tilecode_metrics(from_resolution)

        elif from_dggs == "quadkey":
            tile = mercantile.quadkey_to_tile(from_resolution)
            from_resolution = tile.z
            _, _, from_area = quadkey_metrics(from_resolution)

    except Exception:
        return

    nearest_resolution = None
    min_diff = float("inf")

    if to_dggs == "h3":
        for res in range(16):
            avg_area = h3.average_hexagon_area(res, unit="m^2")
            diff = abs(avg_area - from_area)
            # If the difference is smaller than the current minimum, update the nearest resolution
            if diff < min_diff:
                min_diff = diff
                nearest_resolution = res

    elif to_dggs == "s2":
        for res in range(31):
            _, _, avg_area = s2_metrics(res)
            diff = abs(avg_area - from_area)
            # If the difference is smaller than the current minimum, update the nearest resolution
            if diff < min_diff:
                min_diff = diff
                nearest_resolution = res

    elif to_dggs == "rhealpix":
        for res in range(16):
            _, _, avg_area = rhealpix_metrics(res)
            diff = abs(avg_area - from_area)
            # If the difference is smaller than the current minimum, update the nearest resolution
            if diff < min_diff:
                min_diff = diff
                nearest_resolution = res

    elif to_dggs == "isea4t":
        if platform.system() == "Windows":
            isea4t_dggs = Eaggr(Model.ISEA4T)
            for res in range(26):
                _, _, avg_area, _ = isea4t_metrics(isea4t_dggs, res)
                diff = abs(avg_area - from_area)
                # If the difference is smaller than the current minimum, update the nearest resolution
                if diff < min_diff:
                    min_diff = diff
                    nearest_resolution = res

    elif to_dggs == "qtm":
        for res in range(1, 25):
            _, _, avg_area = qtm_metrics(res)
            diff = abs(avg_area - from_area)
            # If the difference is smaller than the current minimum, update the nearest resolution
            if diff < min_diff:
                min_diff = diff
                nearest_resolution = res

    elif to_dggs == "olc":
        for res in [2, 4, 6, 8, 10, 11, 12, 13, 14, 15]:
            _, _, avg_area = olc_metrics(res)
            diff = abs(avg_area - from_area)
            # If the difference is smaller than the current minimum, update the nearest resolution
            if diff < min_diff:
                min_diff = diff
                nearest_resolution = res

    elif to_dggs == "geohash":
        for res in range(1, 11):
            _, _, avg_area = geohash_metrics(res)
            diff = abs(avg_area - from_area)
            # If the difference is smaller than the current minimum, update the nearest resolution
            if diff < min_diff:
                min_diff = diff
                nearest_resolution = res

    elif to_dggs == "tilecode":
        for res in range(30):
            _, _, avg_area = tilecode_metrics(res)
            diff = abs(avg_area - from_area)
            # If the difference is smaller than the current minimum, update the nearest resolution
            if diff < min_diff:
                min_diff = diff
                nearest_resolution = res

    elif to_dggs == "quadkey":
        for res in range(30):
            _, _, avg_area = quadkey_metrics(res)
            diff = abs(avg_area - from_area)
            # If the difference is smaller than the current minimum, update the nearest resolution
            if diff < min_diff:
                min_diff = diff
                nearest_resolution = res

    return nearest_resolution


def generate_grid(geojson_features, to_dggs, resolution):
    # bbox = extract_bbox(geojson_data)  # Extract bounding box from GeoJSON
    dggs_grid = {}
    if to_dggs == "h3":
        dggs_grid = h3grid.generate_grid_resample(resolution, geojson_features)
    elif to_dggs == "s2":
        dggs_grid = s2grid.generate_grid_resample(resolution, geojson_features)
    elif to_dggs == "rhealpix":
        dggs_grid = rhealpixgrid.generate_grid_resample(resolution, geojson_features)
    elif to_dggs == "isea4t":
        if platform.system() == "Windows":
            isea4t_dggs = Eaggr(Model.ISEA4T)
            dggs_grid = isea4tgrid.generate_grid_resample(
                isea4t_dggs, resolution, geojson_features
            )
    elif to_dggs == "qtm":
        dggs_grid = qtmgrid.generate_grid_resample(resolution, geojson_features)
    elif to_dggs == "olc":
        dggs_grid = olcgrid.generate_grid_resample(resolution, geojson_features)
    elif to_dggs == "geohash":
        dggs_grid = geohashgrid.generate_grid_resample(resolution, geojson_features)
    elif to_dggs == "tilecode":
        dggs_grid = tilecodegrid.generate_grid_resample(resolution, geojson_features)
    elif to_dggs == "quadkey":
        dggs_grid = quadkeygrid.generate_grid_resample(resolution, geojson_features)

    return dggs_grid


def resampling(layer1, layer2, resample_field):
    try:
        layer1_features = []
        for feature in layer1["features"]:
            if resample_field not in feature["properties"]:
                raise ValueError(
                    f"There is no <{resample_field}> field in the input GeoJSON feattures."
                )
            geom = shape(feature["geometry"])
            value = feature["properties"][resample_field]
            layer1_features.append((geom, value))
    except ValueError as e:
        print(e)
        return layer2

    resampled_features = []

    for feature in tqdm(layer2["features"], desc="Resampling", unit=" cells"):
        layer2_shape = shape(feature["geometry"])
        resampled_value = 0
        intersected_parts = []

        for l1_shape, l1_resample in layer1_features:
            if layer2_shape.intersects(l1_shape):
                if not isinstance(l1_resample, Number):
                    print(
                        f"\n Mon-numeric values found in <{resample_field}>. Resampled field calculation failed."
                    )
                    return layer2

                intersection = layer2_shape.intersection(l1_shape)
                intersected_parts.append(intersection)
                if not intersection.is_empty:
                    proportion = intersection.area / l1_shape.area
                    resampled_value += l1_resample * proportion

            if not intersected_parts:
                continue  # Skip features that have no intersection

        # Add resampled fileld to properties
        feature["properties"][resample_field] = round(resampled_value, 3)
        resampled_features.append(feature)

    return {"type": "FeatureCollection", "features": resampled_features}


def main():
    parser = argparse.ArgumentParser(description="DGGS Resample")
    dggs_options = [
        "h3",
        "s2",
        "rhealpix",
        "isea4t",
        "qtm",
        "olc",
        "geohash",
        "tilecode",
        "quadkey",
    ]

    parser.add_argument(
        "-geojson", "--geojson", type=str, required=True, help="Input DGGS"
    )

    parser.add_argument(
        "-fromdggs",
        "--fromdggs",
        type=str,
        choices=dggs_options,
        required=True,
        help="Input DGGS. Choose from: " + ", ".join(dggs_options),
    )
    parser.add_argument("-fromfield", "--fromfield", help="Input DGGS ID field")

    parser.add_argument(
        "-resamplefield",
        "--resamplefield",
        type=str,
        help="Numeric field for resampling",
    )

    parser.add_argument(
        "-todggs",
        "--todggs",
        type=str,
        choices=dggs_options,
        required=True,
        help="Output DGGS",
    )

    parser.add_argument("-r", "--resolution", type=int, help="Output resolution")

    args = parser.parse_args()
    geojson = args.geojson
    from_dggs = args.fromdggs
    from_field = args.fromfield
    resample_field = args.resamplefield
    to_dggs = args.todggs
    to_resolution = args.resolution

    if not os.path.exists(geojson):
        print(f"Error: The file {geojson} does not exist.")
        return

    if from_dggs == to_dggs:
        print("To DGGS must be different with From DGGS")
        return

    with open(geojson) as f:
        geojson_features = json.load(f)

    if not geojson_features["features"]:
        raise ValueError("GeoJSON contains no features.")

    if not to_resolution:
        to_resolution = get_nearest_resolution(
            geojson_features, from_dggs, to_dggs, from_field
        )
    if to_resolution:
        resampled_features = generate_grid(geojson_features, to_dggs, to_resolution)
        if resample_field:
            resampled_features = resampling(
                geojson_features, resampled_features, resample_field
            )

        # Define the GeoJSON file path
        if resampled_features:
            geojson_path = f"{from_dggs}_to_{to_dggs}_{to_resolution}.geojson"
            with open(geojson_path, "w") as f:
                json.dump(resampled_features, f, indent=2)

            print(f"GeoJSON saved as {geojson_path}")
        else:
            print("DGGS Resample failed.")
    else:
        print(
            "There is no appropriate resolutions found in the target DGGS. DGGS Resample failed."
        )


if __name__ == "__main__":
    main()



import argparse
import json
from shapely.geometry import Polygon, box
from tqdm import tqdm
from vgrid.utils.easedggs.constants import grid_spec, ease_crs, geo_crs, levels_specs
from vgrid.utils.easedggs.dggs.grid_addressing import (
    grid_ids_to_geos,
    geo_polygon_to_grid_ids,
)
from vgrid.generator.settings import max_cells, geodesic_dggs_to_feature

# Initialize the geodetic model

geo_bounds = grid_spec["geo"]
min_longitude = geo_bounds["min_x"]
min_lattitude = geo_bounds["min_y"]
max_longitude = geo_bounds["max_x"]
max_latitude = geo_bounds["max_y"]


def get_ease_cells(resolution):
    """
    Generate a list of cell IDs based on the resolution, row, and column.
    """
    n_row = levels_specs[resolution]["n_row"]
    n_col = levels_specs[resolution]["n_col"]

    # Generate list of cell IDs
    cell_ids = []

    # Loop through all rows and columns at the specified resolution
    for row in range(n_row):
        for col in range(n_col):
            # Generate base ID (e.g., L0.RRRCCC for res=0)
            base_id = f"L{resolution}.{row:03d}{col:03d}"

            # Add additional ".RC" for each higher resolution
            cell_id = base_id
            for i in range(1, resolution + 1):
                cell_id += f".{row:1d}{col:1d}"  # For res=1: L0.RRRCCC.RC, res=2: L0.RRRCCC.RC.RC, etc.

            # Append the generated cell ID to the list
            cell_ids.append(cell_id)

    return cell_ids


def get_ease_cells_bbox(resolution, bbox):
    bounding_box = box(*bbox)
    bounding_box_wkt = bounding_box.wkt
    cells_bbox = geo_polygon_to_grid_ids(
        bounding_box_wkt,
        level=resolution,
        source_crs=geo_crs,
        target_crs=ease_crs,
        levels_specs=levels_specs,
        return_centroids=True,
        wkt_geom=True,
    )
    return cells_bbox


def generate_grid(resolution):
    ease_features = []

    level_spec = levels_specs[resolution]
    n_row = level_spec["n_row"]
    n_col = level_spec["n_col"]

    cells = get_ease_cells(resolution)

    for cell in tqdm(
        cells, total=len(cells), desc="Generating EASE DGGS", unit=" cells"
    ):
        geo = grid_ids_to_geos([cell])
        center_lon, center_lat = geo["result"]["data"][0]
        cell_min_lat = center_lat - (180 / (2 * n_row))
        cell_max_lat = center_lat + (180 / (2 * n_row))
        cell_min_lon = center_lon - (360 / (2 * n_col))
        cell_max_lon = center_lon + (360 / (2 * n_col))

        cell_polygon = Polygon(
            [
                [cell_min_lon, cell_min_lat],
                [cell_max_lon, cell_min_lat],
                [cell_max_lon, cell_max_lat],
                [cell_min_lon, cell_max_lat],
                [cell_min_lon, cell_min_lat],
            ]
        )
        if cell_polygon:
            num_edges = 4
            ease_feature = geodesic_dggs_to_feature(
                "ease", str(cell), resolution, cell_polygon, num_edges
            )
            ease_features.append(ease_feature)

    return {"type": "FeatureCollection", "features": ease_features}


def generate_grid_within_bbox(resolution, bbox):
    ease_features = []
    level_spec = levels_specs[resolution]
    n_row = level_spec["n_row"]
    n_col = level_spec["n_col"]

    # Get all grid cells within the bounding box
    cells = get_ease_cells_bbox(resolution, bbox)["result"]["data"]

    if cells:
        for cell in tqdm(cells, desc="Generating EASE DGGS", unit=" cells"):
            geo = grid_ids_to_geos([cell])
            if geo:
                center_lon, center_lat = geo["result"]["data"][0]
                cell_min_lat = center_lat - (180 / (2 * n_row))
                cell_max_lat = center_lat + (180 / (2 * n_row))
                cell_min_lon = center_lon - (360 / (2 * n_col))
                cell_max_lon = center_lon + (360 / (2 * n_col))

                cell_polygon = Polygon(
                    [
                        [cell_min_lon, cell_min_lat],
                        [cell_max_lon, cell_min_lat],
                        [cell_max_lon, cell_max_lat],
                        [cell_min_lon, cell_max_lat],
                        [cell_min_lon, cell_min_lat],
                    ]
                )
                num_edges = 4
                ease_feature = geodesic_dggs_to_feature(
                    "ease", str(cell), resolution, cell_polygon, num_edges
                )
                ease_features.append(ease_feature)

        return {"type": "FeatureCollection", "features": ease_features}


def main():
    parser = argparse.ArgumentParser(description="Generate EASE-DGGS DGGS.")
    parser.add_argument(
        "-r", "--resolution", type=int, required=True, help="resolution [0..6]"
    )
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )

    args = parser.parse_args()
    resolution = args.resolution
    bbox = (
        args.bbox
        if args.bbox
        else [min_longitude, min_lattitude, max_longitude, max_latitude]
    )

    if resolution < 0 or resolution > 6:
        print("Please select a resolution in [0..6] range and try again")
        return

    if bbox == [min_longitude, min_lattitude, max_longitude, max_latitude]:
        level_spec = levels_specs[resolution]
        n_row = level_spec["n_row"]
        n_col = level_spec["n_col"]
        total_cells = n_row * n_col

        print(f"Resolution {resolution} will generate {total_cells} cells ")
        if total_cells > max_cells:
            print(f"which exceeds the limit of {max_cells}.")
            print("Please select a smaller resolution and try again.")
            return
        # Start generating and saving the grid in chunks
        geojson_features = generate_grid(resolution)

    else:
        geojson_features = generate_grid_within_bbox(resolution, bbox)

    if geojson_features:
        # Define the GeoJSON file path
        geojson_path = f"ease_grid_{resolution}.geojson"
        with open(geojson_path, "w") as f:
            json.dump(geojson_features, f, indent=2)

        print(f"GeoJSON saved as {geojson_path}")


if __name__ == "__main__":
    main()



"""
This Python 3.11 code implements ellipsoids of revolution.

- Alexander Raichev (AR), 2012-01-26: Refactored code from release 0.3.

NOTE:

All lengths are measured in meters and all angles are measured in radians
unless indicated otherwise.

By 'ellipsoid' throughout, I mean an ellipsoid of revolution and *not* a general (triaxial) ellipsoid.
Points lying on an ellipsoid are given in geodetic (longitude, latitude) coordinates.
"""

# *****************************************************************************
#       Copyright (C) 2013 Alexander Raichev <alex.raichev@gmail.com>
#
#  Distributed under the terms of the GNU Lesser General Public License (LGPL)
#                  http://www.gnu.org/licenses/
# *****************************************************************************

# Import third-party modules.
import pyproj
from numpy import pi, sqrt, sin, cos, arcsin, arctanh, deg2rad, rad2deg

# Import standard modules.
from random import uniform

# Import my modules.
from vgrid.utils.rhealpixdggs.utils import my_round, auth_lat, auth_rad

# Parameters of some common ellipsoids.
WGS84_A = pyproj.get_ellps_map()["WGS84"]["a"]  # 6378137.0
WGS84_F = 1 / pyproj.get_ellps_map()["WGS84"]["rf"]  # 298.257223563
WGS84_B = WGS84_A * (1 - WGS84_F)
WGS84_E = sqrt(WGS84_F * (2 - WGS84_F))
WGS84_R_A = sqrt(WGS84_A**2 / 2 + WGS84_B**2 / 2 * (arctanh(WGS84_E) / WGS84_E))
R_EM = pyproj.get_ellps_map()["sphere"]["a"]  # 6371000 (Earth's mean radius)


class Ellipsoid(object):
    """
    Represents an ellipsoid of revolution (possibly a sphere) with a
    geodetic longitude-latitude coordinate frame.

    INSTANCE ATTRIBUTES:

    - `sphere` - True if the ellipsoid is a sphere, and False otherwise.
    - `R` - The radius of the ellipsoid in meters, implying that it is a
      sphere.
    - `a` - Major radius of the ellipsoid in meters.
    - `b` - Minor radius of the ellipsoid in meters.
    - `e` - Eccentricity of the ellipsoid.
    - `f` - Flattening of the ellipsoid.
    - `R_A` - Authalic radius of the ellipsoid in meters.
    - `lon_0` - Central meridian.
    - `lat_0` - Latitude of origin.
    - `radians` - If True, use angles measured in radians for all calculations.
      Use degrees otherwise.
    - `phi_0` - The latitude separating the equatorial region and
      the north polar region in the context of the (r)HEALPix projection.

    Except for phi_0, these attribute names match the names of the
    `PROJ ellipsoid parameters <https://proj.org/en/9.4/usage/ellipsoids.html#ellipsoid-spherification-parameters>`
    """

    def __init__(
        self,
        R=None,
        a=WGS84_A,
        b=None,
        e=None,
        f=WGS84_F,
        lon_0=0,
        lat_0=0,
        radians=False,
    ):
        self.lon_0 = lon_0
        self.lat_0 = lat_0
        self.radians = radians
        if R is not None:
            # The ellipsoid is a sphere.
            # Override the other geometric parameters.
            self.sphere = True
            self.R = R
            self.a = R
            self.b = R
            self.e = 0
            self.f = 0
            self.R_A = R
        else:
            self.sphere = False
            self.a = a
            if b is not None:
                # Derive the other geometric parameters from a and b.
                self.b = b
                self.e = sqrt(1 - (b / a) ** 2)
                self.f = (a - b) / a
            elif e is not None:
                # Derive the other geometric parameters from a and e.
                self.e = e
                self.b = a * sqrt(1 - e**2)
                self.f = 1 - sqrt(1 - e**2)
            else:
                self.f = f
                self.b = self.a * (1 - f)
                self.e = sqrt(f * (2 - f))
            self.R_A = auth_rad(self.a, self.e)
        self.phi_0 = auth_lat(arcsin(2.0 / 3), e=self.e, radians=True, inverse=True)
        if not self.radians:
            # Convert to degrees.
            self.phi_0 = rad2deg(self.phi_0)

    def __str__(self):
        result = ["ellipsoid:"]
        # result.append('lengths measured in meters')
        for k, v in sorted(self.__dict__.items()):
            if k == "phi_0":
                continue
            if k in {"sphere", "radians"}:
                result.append("    " + k + " = " + str(v))
            else:
                result.append("    " + k + " = " + str(my_round(v, 15)))
        return "\n".join(result)

    def __eq__(self, other):
        if self.a == other.a and self.b == other.b:
            return True
        else:
            return False

    def __ne__(self, other):
        """
        The inequality relation on cells.
        Since Python 3.3 doesn't automatically create reverse relations
        from given ones, i must define this seemingly redundant relation.
        """
        return not self.__eq__(other)

    def pi(self):
        """
        Return pi if `self.radians` = True and 180 otherwise.
        """
        if self.radians:
            return pi
        else:
            return 180.0

    def random_point(self, lam_min=None, lam_max=None, phi_min=None, phi_max=None):
        """
        Return a point (given in geodetic coordinates) sampled uniformly at
        random from the section of this ellipsoid with longitude in the range
        `lam_min <= lam < lam_max` and latitude in the range
        `phi_min <= phi < phi_max`.
        But avoid the poles.

        EXAMPLES::

           >>> E = UNIT_SPHERE
           >>> print(E.random_point()) # doctest: +SKIP
           (-1.0999574573422948, 0.21029104897701129)

        """
        PI = self.pi()
        if lam_min is None:
            lam_min = -PI
        if lam_max is None:
            lam_max = PI
        if phi_min is None:
            phi_min = -PI / 2
        if phi_max is None:
            phi_max = PI / 2
        if not self.radians:
            # Convert to radians.
            lam_min, lam_max, phi_min, phi_max = deg2rad(
                [lam_min, lam_max, phi_min, phi_max]
            )
        # Pick a longitude.
        while True:
            u = uniform(0, 1)
            lam = (lam_max - lam_min) * u + lam_min
            # Don't include lam_max.
            if lam < lam_max:
                # Success.
                break
        # Pick a latitude.
        delta = pi / 360
        while True:
            v = uniform(0, 1)
            if self.sphere:
                phi = arcsin((sin(phi_max) - sin(phi_min)) * v + sin(phi_min))
            else:
                # Sample from the authalic sphere.
                # The map from the ellipsoid to the authalic sphere is
                # an equiareal diffeomorphism.
                # So a uniform distribution on the authalic sphere gives
                # rise to a uniform distribution on the ellipsoid.
                beta0 = auth_lat(phi_min, e=self.e, radians=True)
                beta1 = auth_lat(phi_max, e=self.e, radians=True)
                beta = arcsin((sin(beta1) - sin(beta0)) * v + sin(beta0))
                phi = auth_lat(beta, e=self.e, radians=True, inverse=True)
            # Avoid the poles.
            if abs(phi) <= pi / 2 - delta:
                # Success.
                break
        if not self.radians:
            # Convert back to degrees.
            lam, phi = rad2deg([lam, phi])
        return lam, phi

    def lattice(self, n=90):
        """
        Return a 2n x n square lattice of longitude-latitude points.

        EXAMPLES::

            >>> E = UNIT_SPHERE
            >>> for p in E.lattice(n=3):
            ...     print(p)
            (-150.0, -60.0)
            (-150.0, 0.0)
            (-150.0, 60.0)
            (-90.0, -60.0)
            (-90.0, 0.0)
            (-90.0, 60.0)
            (-30.0, -60.0)
            (-30.0, 0.0)
            (-30.0, 60.0)
            (30.0, -60.0)
            (30.0, 0.0)
            (30.0, 60.0)
            (90.0, -60.0)
            (90.0, 0.0)
            (90.0, 60.0)
            (150.0, -60.0)
            (150.0, 0.0)
            (150.0, 60.0)

        """
        PI = self.pi()
        # Longitudinal and latitudinal spacing between points.
        delta = PI / n
        return [
            (-PI + delta * (0.5 + i), -PI / 2 + delta * (0.5 + j))
            for i in range(2 * n)
            for j in range(n)
        ]

    def meridian(self, lam, n=200):
        """
        Return a list of `n` equispaced longitude-latitude
        points lying along the meridian of longitude `lam`.
        Avoid the poles.
        """
        PI = self.pi()
        delta = PI / n
        return [(lam, -PI / 2 + delta * (0.5 + i)) for i in range(n)]

    def parallel(self, phi, n=200):
        """
        Return a list of `2*n` equispaced longitude-latitude
        points lying along the parallel of latitude `phi`.
        """
        PI = self.pi()
        delta = PI / n
        return [(-PI + delta * (0.5 + i), phi) for i in range(2 * n)]

    def graticule(self, n=400, spacing=None):
        """
        Return a list of longitude-latitude points sampled from a
        longitude-latitude graticule on this ellipsoid with the given
        spacing between meridians and between parallels.
        The number of points on longitude and latitude per pi radians is `n`.
        The spacing should be specified in the angle units used for this
        ellipsoid.
        If `spacing=None`, then a default spacing of pi/16 radians will be set.

        EXAMPLES::

            >>> E = UNIT_SPHERE
            >>> print(len(E.graticule(n=400)))
            25600

        """
        PI = self.pi()
        result = []
        # delta = PI/n
        # Set default spacing.
        if spacing is None:
            spacing = PI / 16
        # Longitude lines.
        lam = -PI
        while lam < PI:
            # result.extend([(lam, -PI/2 + delta*(0.5 + i)) for i in range(n)])
            result.extend(self.meridian(lam, n))
            lam += spacing
        # Latitude lines. Avoid the poles.
        eps = PI / 360
        phi = -PI / 2 + eps
        while phi < PI / 2:
            # result.extend([(-PI + delta*(0.5 + i), phi) for i in range(2*n)])
            result.extend(self.parallel(phi, n))
            phi += spacing
        return result

    def get_points(self, filename):
        """
        Return a list of longitude-latitude points contained in
        the file with filename `filename`.
        Assume the file is a text file containing at most one
        longitude-latitude point per line with the coordinates separated by
        whitespace and angles given in degrees.
        """
        result = []
        for line in open(filename, "rb"):
            if line[0] not in ["-", "1", "2", "3", "4", "5", "6", "7", "8", "9"]:
                # Ignore line.
                continue
            else:
                # Split coordinate pair on whitespace.
                p = [float(x) for x in line.split()]
                result.append(p)
        if self.radians:
            # Convert to radians.
            result = [deg2rad(p) for p in result]
        return result

    def xyz(self, lam, phi):
        """
        Given a point on this ellipsoid with longitude-latitude coordinates
        `(lam, phi)`, return the point's 3D rectangular coordinates.

        EXAMPLES::

            >>> E = UNIT_SPHERE
            >>> print(tuple(x.tolist() for x in my_round(E.xyz(0, 45), 15)))
            (0.707106781186548, 0.0, 0.707106781186548)

        """
        a = self.a
        e = self.e
        if not self.radians:
            lam, phi = deg2rad([lam, phi])
        # Equals a iff e = 0 (sphere):
        N = a / sqrt(1 - e**2 * sin(phi) ** 2)
        return (
            N * cos(lam) * cos(phi),
            N * sin(lam) * cos(phi),
            N * (1 - e**2) * sin(phi),
        )


# Define some common ellipsoids.
WGS84_ELLIPSOID = Ellipsoid(a=WGS84_A, f=WGS84_F)
WGS84_ELLIPSOID_RADIANS = Ellipsoid(a=WGS84_A, f=WGS84_F, radians=True)
WGS84_ASPHERE = Ellipsoid(R=WGS84_R_A)
WGS84_ASPHERE_RADIANS = Ellipsoid(R=WGS84_R_A, radians=True)
EMR_SPHERE = Ellipsoid(R=R_EM)
EMR_SPHERE_RADIANS = Ellipsoid(R=R_EM, radians=True)
UNIT_SPHERE = Ellipsoid(R=1)
UNIT_SPHERE_RADIANS = Ellipsoid(R=1, radians=True)



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

import os
import sys
import platform

from ctypes import (
    cdll,
    CDLL,
    c_void_p,
    c_char_p,
    c_ushort,
    c_int,
    c_double,
    POINTER,
    byref,
    create_string_buffer,
    cast,
    c_bool,
)

# Top level of the namespace is different in this file in Python 2
if sys.version_info < (3, 0):
    from exceptions.eaggr_exception import EaggrException
    from enums.dggs_return_code import DggsReturnCode
    from enums.model import check_model
    from enums.shape_string_format import check_shape_string_format
    from shapes.native_structures import (
        get_LAT_LONG_SHAPE_array,
        get_LAT_LONG_POINT_array,
        LAT_LONG_POINT,
        LAT_LONG_SHAPE,
        get_DGGS_CELL_array,
        DGGS_CELL,
        DGGS_SHAPE,
    )
else:
    from vgrid.utils.eaggr.exceptions.eaggr_exception import EaggrException
    from vgrid.utils.eaggr.enums.dggs_return_code import DggsReturnCode
    from vgrid.utils.eaggr.enums.model import check_model
    from vgrid.utils.eaggr.enums.shape_string_format import check_shape_string_format
    from vgrid.utils.eaggr.shapes.native_structures import (
        get_LAT_LONG_SHAPE_array,
        get_LAT_LONG_POINT_array,
        LAT_LONG_POINT,
        LAT_LONG_SHAPE,
        get_DGGS_CELL_array,
        DGGS_CELL,
        DGGS_SHAPE,
    )


## Python API for the EAGGR library.
class Eaggr(object):
    ## Number of characters in EAGGR DLL's version string (including the null
    #  terminating character).
    _EAGGR_VERSION_STRING_LENGTH = 5

    ## Maximum number of parents a DGGS cell can have.
    _EAGGR_MAX_PARENT_CELLS = 3
    ## Maximum number of children a DGGS cell can have.
    _EAGGR_MAX_CHILD_CELLS = 7
    ## Maximum number of siblings a DGGS cell can have.
    _EAGGR_MAX_SIBLING_CELLS = 15

    ## Handle to the Proj.4 library DLL.
    _proj_dll_handle = None
    ## Handle to the GDAL library DLL.
    _gdal_dll_handle = None
    ## Handle to the EAGGR DLL.
    _eaggr_dll_handle = None
    ## Reference to the ctypes library object. Used to call the functions in the EAGGR DLL.
    _eaggr_dll = None

    ## Loads the necessary DLLs and opens a handle to the EAGGR library.
    #  @param dggs_model Specifies the type of DGGS model to use (enums.Model).
    #  @throw EaggrException Thrown if unable to open a handle to the EAGGR library.
    #  @throw IOError Thrown if one of the DLLs cannot be found.
    def __init__(self, dggs_model):
        # Check input argument
        check_model(dggs_model)
        # Open the DLLs if they are not already open
        if Eaggr._eaggr_dll is None:
            self._open_dlls()
        # Open a handle to the DGGS model
        self._dggs_handle = c_void_p()
        func = getattr(Eaggr._eaggr_dll, "EAGGR_OpenDggsHandle")
        func.argtypes = [c_int, POINTER(c_void_p)]
        func.restype = c_int
        return_code = func(dggs_model, byref(self._dggs_handle))
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())

    ## Closes the handle to the model and frees the resources allocated to the DLLs.
    def __del__(self):
        if Eaggr._eaggr_dll is not None and self._dggs_handle is not None:
            # Call the DLL to close the handle to the DGGS
            func = getattr(Eaggr._eaggr_dll, "EAGGR_CloseDggsHandle")
            func.argtypes = [c_void_p]
            func.restype = c_int
            # Return code is ignored - cannot do anything about it in a destructor
            func(self._dggs_handle)

    ## Gets the version of the DLL as a string.
    #  @return Version of the DLL as a string.
    #  @throw EaggrException Thrown if unable to get the version number.
    def get_version(self):
        # Set up the arguments to the DLL function
        version = create_string_buffer(Eaggr._EAGGR_VERSION_STRING_LENGTH)
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_GetVersion")
        func.argtypes = [c_char_p]
        func.restype = c_int
        return_code = func(version)
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(
                return_code, "Failed to get version number of EAGGR DLL"
            )
        return version.value.decode("utf-8")

    ## Converts a single point in lat / long coordinates into a DGGS cell.
    #  @param point Point defined by lat / long coordinates.
    #  @return DGGS cell.
    #  @throw EaggrException Thrown if unable to convert the lat / long point.
    def convert_point_to_dggs_cell(self, point):
        # Set up the arguments to the DLL function (create array of one point)
        lat_long_points = get_LAT_LONG_POINT_array([point])
        no_of_points = c_ushort(1)
        output_cells = (DGGS_CELL * no_of_points.value)()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_ConvertPointsToDggsCells")
        func.argtypes = [
            c_void_p,
            POINTER(LAT_LONG_POINT),
            c_ushort,
            POINTER(DGGS_CELL),
        ]
        func.restype = c_int
        return_code = func(
            self._dggs_handle, lat_long_points, no_of_points, output_cells
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Return the first (and only) DGGS cell in the list
        return output_cells[0].to_dggs_cell()

    ## Converts a list of shapes in lat / long coordinates into a list of
    #  shapes defined by DGGS cells.
    #  @param shapes List of shapes defined by lat / long coordinates.
    #  @return List of DGGS shapes, with the same number of elements as the input list.
    #  @throw EaggrException Thrown if unable to convert the lat / long shapes.
    def convert_shapes_to_dggs_shapes(self, shapes):
        # Set up the arguments to the DLL function
        lat_long_shapes = get_LAT_LONG_SHAPE_array(shapes)
        no_of_shapes = c_ushort(len(shapes))
        output_shapes = POINTER(DGGS_SHAPE)()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_ConvertShapesToDggsShapes")
        func.argtypes = [
            c_void_p,
            POINTER(LAT_LONG_SHAPE),
            c_ushort,
            POINTER(POINTER(DGGS_SHAPE)),
        ]
        func.restype = c_int
        return_code = func(
            self._dggs_handle, lat_long_shapes, no_of_shapes, byref(output_shapes)
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Process the output data
        dggs_shapes = []
        for shape_index in range(0, no_of_shapes.value):
            dggs_shapes.append(output_shapes[shape_index].to_dggs_shape())
        # Deallocate memory for the output data
        self._deallocate_dggs_shapes(output_shapes, no_of_shapes.value)
        # Return the array of DGGS shapes
        return dggs_shapes

    ## Converts a shape string into a list of shapes defined by DGGS cells.
    #  @param string String containing shape information in one of the supported formats.
    #  @param string_format Format used for the string (enums.ShapeStringFormat).
    #  @param accuracy Defines an area of accuracy (in metres squared) to use for every point in the string.
    #  @return List containing the input shapes defined by DGGS cells.
    #  @throw EaggrException Thrown if unable to convert the shape string.
    def convert_shape_string_to_dggs_shapes(self, string, string_format, accuracy):
        # Check requested string format is valid
        check_shape_string_format(string_format)
        # Set up the arguments to the DLL function
        try:
            shape_string = cast(string.encode("utf-8"), c_char_p).value
        except AttributeError:
            raise ValueError("Invalid shape string")
        output_shapes = POINTER(DGGS_SHAPE)()
        no_of_shapes = c_ushort()
        try:
            input_accuracy = c_double(accuracy)
        except TypeError:
            raise ValueError("Invalid accuracy value")
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_ConvertShapeStringToDggsShapes")
        func.argtypes = [
            c_void_p,
            c_char_p,
            c_int,
            c_double,
            POINTER(POINTER(DGGS_SHAPE)),
            POINTER(c_ushort),
        ]
        func.restype = c_int
        return_code = func(
            self._dggs_handle,
            shape_string,
            string_format,
            input_accuracy,
            byref(output_shapes),
            byref(no_of_shapes),
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Process the output data
        dggs_shapes = []
        for shape_index in range(0, no_of_shapes.value):
            dggs_shapes.append(output_shapes[shape_index].to_dggs_shape())
        # Deallocate memory for output data
        self._deallocate_dggs_shapes(output_shapes, no_of_shapes.value)
        # Return the array of DGGS shapes
        return dggs_shapes

    ## Converts a single DGGS cell into a point with lat / long coordinates.
    #  @param cell DGGS cell.
    #  @return Lat / long point.
    #  @throw EaggrException Thrown if unable to convert the DGGS cell.
    def convert_dggs_cell_to_point(self, cell):
        # Call convert_dggs_cells_to_points() with just one cell
        lat_long_points = self.convert_dggs_cells_to_points([cell])
        return lat_long_points[0]

    ## Converts a list of DGGS cells into a list of points with lat / long coordinates.
    #  @param cells List of DGGS cells.
    #  @return List of lat / long points.
    #  @throw EaggrException Thrown if unable to convert the DGGS cells.
    def convert_dggs_cells_to_points(self, cells):
        # Set up the arguments to the DLL function
        dggs_cells = get_DGGS_CELL_array(cells)
        no_of_cells = c_ushort(len(cells))
        output_points = (LAT_LONG_POINT * no_of_cells.value)()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_ConvertDggsCellsToPoints")
        func.argtypes = [
            c_void_p,
            POINTER(DGGS_CELL),
            c_ushort,
            POINTER(LAT_LONG_POINT),
        ]
        func.restype = c_int
        return_code = func(self._dggs_handle, dggs_cells, no_of_cells, output_points)
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Process the output data
        lat_long_points = []
        for point_index in range(0, no_of_cells.value):
            lat_long_points.append(output_points[point_index].to_lat_long_point())
        # Return the array of DGGS shapes
        return lat_long_points

    ## Converts a list of DGGS cells into a string of lat / long points.
    #  @param cells List of DGGS cells.
    #  @param string_format Format to use for the output string (enums.ShapeStringFormat).
    #  @return String defining the DGGS cells in lat / long coordinates.
    #  @throw EaggrException Thrown if unable to convert the DGGS cells.
    def convert_dggs_cells_to_shape_string(self, cells, string_format):
        # Check requested string format is valid
        check_shape_string_format(string_format)
        # Set up the arguments to the DLL function
        dggs_cells = get_DGGS_CELL_array(cells)
        no_of_cells = c_ushort(len(cells))
        output_string = c_char_p()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_ConvertDggsCellsToShapeString")
        func.argtypes = [
            c_void_p,
            POINTER(DGGS_CELL),
            c_ushort,
            c_int,
            POINTER(c_char_p),
        ]
        func.restype = c_int
        return_code = func(
            self._dggs_handle,
            dggs_cells,
            no_of_cells,
            string_format,
            byref(output_string),
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Store the string from native memory
        returnString = cast(output_string, c_char_p).value.decode("utf-8")
        # Free the native string memory
        func = getattr(Eaggr._eaggr_dll, "EAGGR_DeallocateString")
        func.argtypes = [c_void_p, POINTER(c_char_p)]
        func.restype = c_int
        return_code = func(self._dggs_handle, byref(output_string))
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Return the shape string
        return returnString

    ## Converts a cell area to a shape string containing a lat / long polygon.
    #  @param cell DGGS cell to get the shape string for.
    #  @param string_format Format to use for the output string (enums.ShapeStringFormat).
    #  @return String defining the DGGS cell shape.
    #  @throw EaggrException Thrown if unable to get the DGGS cell shape.
    def convert_dggs_cell_outline_to_shape_string(self, cell, string_format):
        # Check requested string format is valid
        check_shape_string_format(string_format)
        # Set up the arguments to the DLL function
        dggs_cell = DGGS_CELL()
        dggs_cell.from_dggs_cell(cell)
        output_string = c_char_p()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_ConvertDggsCellOutlineToShapeString")
        func.argtypes = [c_void_p, DGGS_CELL, c_int, POINTER(c_char_p)]
        func.restype = c_int
        return_code = func(
            self._dggs_handle, dggs_cell, string_format, byref(output_string)
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Store the string from native memory
        returnString = cast(output_string, c_char_p).value.decode("utf-8")
        # Free the native string memory
        func = getattr(Eaggr._eaggr_dll, "EAGGR_DeallocateString")
        func.argtypes = [c_void_p, POINTER(c_char_p)]
        func.restype = c_int
        return_code = func(self._dggs_handle, byref(output_string))
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Return the shape string
        return returnString

    ## Outputs the parents of the specified cell.
    #
    #  Parent cells are defined as the cells in the resolution below which share
    #  some or all of the area of the specified cell.
    #  @param cell DGGS cell to find the parents of.
    #  @return Parent cells.
    #  @throw  EaggrException Thrown if unable to get the parent cells.
    def get_dggs_cell_parents(self, cell):
        # Set up the arguments to the DLL function
        dggs_cell = DGGS_CELL()
        dggs_cell.from_dggs_cell(cell)
        output_cells = (DGGS_CELL * Eaggr._EAGGR_MAX_PARENT_CELLS)()
        no_of_parents = c_ushort()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_GetDggsCellParents")
        func.argtypes = [c_void_p, DGGS_CELL, POINTER(DGGS_CELL), POINTER(c_ushort)]
        func.restype = c_int
        return_code = func(
            self._dggs_handle, dggs_cell, output_cells, byref(no_of_parents)
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Process the output data
        parents_cells = []
        for cell_index in range(0, no_of_parents.value):
            parents_cells.append(output_cells[cell_index].to_dggs_cell())
        # Return the array of parent cells
        return parents_cells

    ## Outputs the children of the specified cell.
    #
    #  Child cells are defined as the cells in the resolution above which share
    #  some or all of the area of the specified cell.
    #  @param cell DGGS cell to find the children of.
    #  @return Child cells.
    #  @throw EaggrException Thrown if unable to get the parent cells.
    def get_dggs_cell_children(self, cell):
        # Set up the arguments to the DLL function
        dggs_cell = DGGS_CELL()
        dggs_cell.from_dggs_cell(cell)
        output_cells = (DGGS_CELL * Eaggr._EAGGR_MAX_CHILD_CELLS)()
        no_of_children = c_ushort()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_GetDggsCellChildren")
        func.argtypes = [c_void_p, DGGS_CELL, POINTER(DGGS_CELL), POINTER(c_ushort)]
        func.restype = c_int
        return_code = func(
            self._dggs_handle, dggs_cell, output_cells, byref(no_of_children)
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Process the output data
        child_cells = []
        for cell_index in range(0, no_of_children.value):
            child_cells.append(output_cells[cell_index].to_dggs_cell())
        # Return the array of children cells
        return child_cells

    ## Outputs the siblings of the specified cell.
    #
    #  Siblings cells are defined as those cells in the same resolution that share
    #  area with the cells in the resolution above where these share area with the
    #  specified cell.
    #  @param cell DGGS cell to find the siblings of.
    #  @return Sibling cells.
    #  @throw  EaggrException Thrown if unable to get the sibling cells.
    def get_dggs_cell_siblings(self, cell):
        # Set up the arguments to the DLL function
        dggs_cell = DGGS_CELL()
        dggs_cell.from_dggs_cell(cell)
        output_cells = (DGGS_CELL * Eaggr._EAGGR_MAX_SIBLING_CELLS)()
        no_of_siblings = c_ushort()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_GetDggsCellSiblings")
        func.argtypes = [c_void_p, DGGS_CELL, POINTER(DGGS_CELL), POINTER(c_ushort)]
        func.restype = c_int
        return_code = func(
            self._dggs_handle, dggs_cell, output_cells, byref(no_of_siblings)
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Process the output data
        siblings_cells = []
        for cell_index in range(0, no_of_siblings.value):
            siblings_cells.append(output_cells[cell_index].to_dggs_cell())
        # Return the array of sibling cells
        return siblings_cells

    ## Outputs the highest resolution cell that contains all the given cells.
    #  @param cells List of DGGS cells to find the bounding cell of.
    #  @return Cell containing all of the cells in the input list.
    #  @throw EaggrException Thrown if unable to get the bounding cell.
    def get_bounding_dggs_cell(self, cells):
        # Set up the arguments to the DLL function
        dggs_cells = get_DGGS_CELL_array(cells)
        no_of_cells = c_ushort(len(cells))
        output_cell = DGGS_CELL()
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_GetBoundingDggsCell")
        func.argtypes = [c_void_p, POINTER(DGGS_CELL), c_ushort, POINTER(DGGS_CELL)]
        func.restype = c_int
        return_code = func(
            self._dggs_handle, dggs_cells, no_of_cells, byref(output_cell)
        )
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())
        # Return the array of DGGS shapes
        return output_cell.to_dggs_cell()

    ## Creates a KML file for displaying the given cells on mapping applications,
    #  such as Google Earth.
    #  @param filename Filename of the KML file to be created.
    #  @param cells List of DGGS cells to include in the KML file.
    #  @throw EaggrException Thrown if unable to create the KML file.
    def create_dggs_kml_file(self, filename, cells):
        # Set up the arguments to the DLL function
        try:
            kml_filename = cast(filename.encode("utf-8"), c_char_p).value
        except AttributeError:
            raise ValueError("Invalid filename for KML file")
        dggs_cells = get_DGGS_CELL_array(cells)
        no_of_cells = c_ushort(len(cells))
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_CreateDggsKmlFile")
        func.argtypes = [c_void_p, c_char_p, POINTER(DGGS_CELL), c_ushort]
        func.restype = c_int
        return_code = func(self._dggs_handle, kml_filename, dggs_cells, no_of_cells)
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())

    ## Opens handles to the DGGS DLL and its dependencies.
    #  @note Because the DLLs used by the EAGGR library are not in the run directory or the path
    #        they must be manually loaded beforehand so the EAGGR library DLL can find them.
    #  @throw IOError Thrown if unable to open a handle to one of the DLLs.
    def _open_dlls(self):
        try:  # to handle the error in Linux
            # DLL directories and filenames depend on the platform
            if platform.system() == "Windows":
                proj_library_name = "libproj-9.dll"
                gdal_library_name = "libgdal-20.dll"
                dggs_library_name = "eaggr.dll"
                if platform.architecture()[0] == "64bit":
                    gcc_library_name = "libgcc_s_seh-1.dll"
                    dll_dir = os.path.join(
                        os.path.dirname(__file__), "dlls", "win32-x86-64"
                    )
                else:
                    gcc_library_name = "libgcc_s_dw2-1.dll"
                    dll_dir = os.path.join(
                        os.path.dirname(__file__), "dlls", "win32-x86"
                    )
            else:
                proj_library_name = "libproj.so"
                gdal_library_name = "libgdal.so"
                dggs_library_name = "libeaggr.so"
                dll_dir = os.path.join(
                    os.path.dirname(__file__), "dlls", "linux-x86-64"
                )
            # Check the chosen DLL directory exists
            if not os.path.exists(dll_dir):
                raise IOError("Cannot find DLL directory - " + dll_dir)
            # Load MinGW DLLs in the correct order (required on Windows only)
            if platform.system() == "Windows":
                thread_path = os.path.join(dll_dir, "libwinpthread-1.dll")
                if not os.path.isfile(thread_path):
                    raise IOError(
                        "Cannot find the Windows thread library - " + thread_path
                    )
                Eaggr._proj_dll_handle = cdll.LoadLibrary(thread_path)
                gcc_path = os.path.join(dll_dir, gcc_library_name)
                if not os.path.isfile(gcc_path):
                    raise IOError("Cannot find the GCC library - " + gcc_path)
                Eaggr._proj_dll_handle = cdll.LoadLibrary(gcc_path)
                stdc_path = os.path.join(dll_dir, "libstdc++-6.dll")
                if not os.path.isfile(stdc_path):
                    raise IOError("Cannot find the std C++ library - " + stdc_path)
                Eaggr._proj_dll_handle = cdll.LoadLibrary(stdc_path)
            # Load Proj4 and GDAL library DLLs
            proj_path = os.path.join(dll_dir, proj_library_name)
            if not os.path.isfile(proj_path):
                raise IOError("Cannot find the Proj4 library - " + proj_path)
            Eaggr._proj_dll_handle = cdll.LoadLibrary(proj_path)
            gdal_path = os.path.join(dll_dir, gdal_library_name)
            if not os.path.isfile(gdal_path):
                raise IOError("Cannot find the GDAL library - " + gdal_path)
            Eaggr._gdal_dll_handle = cdll.LoadLibrary(gdal_path)
            # Now load the EAGGR DLL
            dggs_path = os.path.join(dll_dir, dggs_library_name)
            if not os.path.isfile(dggs_path):
                raise IOError("Cannot find the EAGGR library - " + dggs_path)
            Eaggr._eaggr_dll_handle = cdll.LoadLibrary(dggs_path)
            Eaggr._eaggr_dll = CDLL(dggs_path)
            if Eaggr._eaggr_dll is None:
                raise IOError("Unable to open handle to EAGGR DLL")
        except:
            return

    ## Deallocates the memory used by the DGGS shapes that are output from the DLL.
    #  @param output_shapes Array containing the DGGS shapes.
    #  @param no_of_shapes Number of shapes in the array.
    #  @throw EaggrException Thrown if unable to deallocate the memory for the DGGS shapes.
    def _deallocate_dggs_shapes(self, output_shapes, no_of_shapes):
        # Configure and call the DLL function
        func = getattr(Eaggr._eaggr_dll, "EAGGR_DeallocateDggsShapes")
        func.argtypes = [c_void_p, POINTER(POINTER(DGGS_SHAPE)), c_ushort]
        func.restype = c_int
        return_code = func(self._dggs_handle, byref(output_shapes), no_of_shapes)
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, self._get_last_error_message())

    ## Outputs a message for the last error to occur.
    #  @return Error message string.
    #  @throw EaggrException Thrown if unable to get the last error message.
    def _get_last_error_message(self):
        func = getattr(Eaggr._eaggr_dll, "EAGGR_GetLastErrorMessage")
        func.argtypes = [c_void_p, POINTER(c_char_p), POINTER(c_ushort)]
        func.restype = c_int
        error_message = c_char_p()
        message_length = c_ushort()
        return_code = func(
            self._dggs_handle, byref(error_message), byref(message_length)
        )
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException("Error occurred, but failed to get error message")
        # Store the string from native memory
        returnString = cast(error_message, c_char_p).value.decode("utf-8")
        # Free the native string memory
        func = getattr(Eaggr._eaggr_dll, "EAGGR_DeallocateString")
        func.argtypes = [c_void_p, POINTER(c_char_p)]
        func.restype = c_int
        return_code = func(self._dggs_handle, byref(error_message))
        # Check the return code
        if return_code != DggsReturnCode.DGGS_SUCCESS:
            raise EaggrException(return_code, "Failed to free native string memory.")
        # Return the shape string
        return returnString

    ## Outputs the result of a spatial analysis between two shapes.
    #
    #  Two DGGS_Shapes are compared using a DGGS_AnalysisType spatial analysis type.
    #  @param baseShape The base DGGSShape object used in the spatial analysis
    #  @param comparisonShape The comparison DGGSShape object used in the spatial analysis
    #  @param analysisType The spatial analysis type, taken from the dggs_shape_type enum
    #  @return flag indicating result of spatial analysis
    #  @throw  EaggrException Thrown if unable to get shape comparison
    def compare_dggs_shapes(self, baseShape, comparisonShape, analysisType):
        try:
            # Set up the arguments to the DLL function
            base_dggs_shape = DGGS_SHAPE()
            comparison_dggs_shape = DGGS_SHAPE()
            dggs_shape_analysis_type = analysisType
            base_dggs_shape.from_dggs_shape(baseShape)
            comparison_dggs_shape.from_dggs_shape(comparisonShape)
            func_output = c_bool()
            # Configure and call the DLL function
            func = getattr(Eaggr._eaggr_dll, "EAGGR_CompareShapes")
            func.argtypes = [
                c_void_p,
                c_int,
                POINTER(DGGS_SHAPE),
                POINTER(DGGS_SHAPE),
                POINTER(c_bool),
            ]
            func.restype = c_int
            return_code = func(
                self._dggs_handle,
                dggs_shape_analysis_type,
                base_dggs_shape,
                comparison_dggs_shape,
                byref(func_output),
            )
            output_boolean = func_output.value
            # Check the return code
            if return_code != DggsReturnCode.DGGS_SUCCESS:
                raise EaggrException(return_code, self._get_last_error_message())
            # Return the boolean indicating the result of the shape comparison
            return output_boolean

        except TypeError:
            raise ValueError(
                "Analysis type must be CONTAINS, COVERED_BY, COVERS, CROSSES, DISJOINT, "
                "EQUALS, INTERSECTS, OVERLAPS, TOUCHES or  WITHIN"
            )



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

from setuptools import setup
from codecs import open
from os import path
from _ast import __version__

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "DESCRIPTION.rst"), encoding="utf-8") as f:
    long_description = f.read()
with open(path.join(here, "eaggr", "version.py"), encoding="utf-8") as f:
    exec(f.read())

setup(
    name="eaggr",
    version=__version__,
    description="EAGGR Python Library",
    long_description=long_description,
    url="http://www.riskaware.co.uk",
    author="Riskaware Ltd",
    author_email="ian.bush@riskaware.co.uk",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
    ],
    keywords=[
        "eaggr",
        "dggs",
    ],
    packages=[
        "eaggr",
        "eaggr.enums",
        "eaggr.exceptions",
        "eaggr.shapes",
        "eaggr.dlls.win32-x86",
        "eaggr.dlls.win32-x86-64",
        "eaggr.dlls.linux-x86-64",
    ],
    install_requires=[],
    extras_require={},
    package_data={
        "eaggr.dlls.win32-x86": ["*.dll"],
        "eaggr.dlls.win32-x86-64": ["*.dll"],
        "eaggr.dlls.linux-x86-64": ["*.so*"],
    },
    data_files=[],
    entry_points={},
)
